###############################################################################
# (c) Copyright 2024 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "COPYING".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
from abc import ABCMeta, abstractmethod

from DIRAC.Core.Utilities.ReturnValues import returnValueOrRaise
from LHCbDIRAC.TransformationSystem.Client.TransformationClient import TransformationClient
from LHCbDIRAC.ProductionManagementSystem.Client.ProductionRequestClient import ProductionRequestClient


class Action(metaclass=ABCMeta):
    # Should this action be ran without prompting the user?
    safe = False

    @abstractmethod
    def message(self): ...

    @abstractmethod
    def run(self): ...


class UpdateRequestStateAction(Action):
    def __init__(self, request_id, state):
        self.request_id = request_id
        self.state = state

    def message(self):
        return f"{self.__class__.__name__}({self.request_id}, {self.state})"

    def run(self):
        returnValueOrRaise(
            ProductionRequestClient().updateProductionRequest(int(self.request_id), {"RequestState": self.state})
        )
        print(f"Request {self.request_id} set to {self.state}")


class ExtendEndRunAction(Action):
    def __init__(self, issue, transform_id, end_run):
        self.issue = issue
        self.transform_id = transform_id
        self.new_end_run = end_run

    def message(self):
        return f"{self.__class__.__name__}({self.transform_id}, {self.new_end_run})"

    def run(self):
        returnValueOrRaise(TransformationClient().setBookkeepingQueryEndRun(self.transform_id, self.new_end_run))


class UpdateIssueStateLabel(Action):
    def __init__(self, issue, old_state, new_state):
        self.issue = issue
        self.old_state = old_state
        self.new_state = new_state

    def message(self):
        return f"{self.__class__.__name__}({self.issue}, {self.new_state})"

    def run(self):
        self.issue.labels.pop(self.issue.labels.index(f"state::{self.old_state}"))
        self.issue.labels.append(f"state::{self.new_state}")
        self.issue.issue.save()
        print(f"Issue {self.issue} set to {self.new_state}")


class SubmitProductionsAction(Action):
    def __init__(self, issue):
        self.issue = issue

    def message(self):
        return f"{self.__class__.__name__}({self.issue}, validation={'state::ready' not in self.issue.labels})"

    def run(self):
        from .launching import start_productions
        from .integrations import OperationsLogbook

        logbook = OperationsLogbook()
        start_productions(logbook, self.issue)


class SubmitRemovalTransformationAction(Action):
    def __init__(self, issue, request_id: int):
        self.issue = issue
        self.request_id = request_id

    def message(self):
        return f"{self.__class__.__name__}({self.issue})"

    def run(self):
        from .launching import start_removal_transformation
        from .integrations import OperationsLogbook

        logbook = OperationsLogbook()
        start_removal_transformation(logbook, self.issue, self.request_id)


class CheckValidationAction(Action):
    def __init__(self, issue, request_id):
        self.issue = issue
        self.request_id = request_id

    def message(self):
        return f"{self.__class__.__name__}({self.issue}, {self.request_id})"

    def run(self):
        from .launching import check_validation

        check_validation(self.issue, self.request_id)


class UpdateTransformationStatusAction(Action):
    def __init__(self, transform_id, status):
        self.transform_id = transform_id
        self.status = status

    def message(self):
        return f"{self.__class__.__name__}({self.transform_id}, {self.status})"

    def run(self):
        tc = TransformationClient()
        returnValueOrRaise(tc.setTransformationParameter(self.transform_id, "Status", self.status))
        print(f"Transformation {self.transform_id} set to {self.status}")


class CleanValidationAction(Action):
    def __init__(self, issue, request_id):
        self.issue = issue
        self.request_id = request_id

    def message(self):
        return f"{self.__class__.__name__}({self.issue}, {self.request_id})"

    def run(self):
        tc = TransformationClient()
        for tid in self.issue.metadata.validations[self.request_id].transform_ids:
            # Safety check
            if returnValueOrRaise(tc.getAdditionalParameters(tid))["configName"] != "validation":
                raise ValueError(f"Transformation {tid} is not a validation")
            returnValueOrRaise(tc.setTransformationParameter(tid, "Status", "Cleaning"))
            print(f"Transformation {tid} set to Cleaning")

        self.issue.metadata.validations[self.request_id].cleaned = True
        self.issue.update_metadata()


class CompleteTransformationAction(Action):
    def __init__(self, issue, transform_id):
        self.issue = issue
        self.transform_id = transform_id

    def message(self):
        return f"{self.__class__.__name__}({self.issue}, {self.transform_id})"

    def run(self):
        tc = TransformationClient()
        returnValueOrRaise(tc.setTransformationParameter(self.transform_id, "Status", "Completed"))
        print(f"Transformation {self.transform_id} set to Completed")


class CloseIssueAction(Action):
    def __init__(self, issue):
        self.issue = issue

    def message(self):
        return f"{self.__class__.__name__}({self.issue})"

    def run(self):
        self.issue.issue.state_event = "close"
        self.issue.issue.save()
        print(f"Issue {self.issue} closed")


class AddLabelAction(Action):
    safe = True

    def __init__(self, repo, issue, label):
        self.repo = repo
        self.issue = issue
        self.label = label

    def message(self):
        return f"{self.__class__.__name__}({self.issue}, {self.label})"

    def run(self):
        if self.label not in self.repo.known_labels:
            prefix, _ = self.label.split("::")
            for project_label, color in self.repo.known_labels.items():
                if project_label.startswith(prefix):
                    break
            else:
                raise ValueError(f"Unknown label prefix {prefix}")
            self.repo.project.labels.create({"name": self.label, "color": color})
            # Clear the cache
            del self.repo.known_labels
        self.issue.labels.append(self.label)
        self.issue.issue.save()



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""An agent to extend MC productions based on the remaining events to
produce."""
import math
import datetime

from DIRAC import S_OK, S_ERROR
from DIRAC.ConfigurationSystem.Client.Helpers.Operations import Operations
from DIRAC.TransformationSystem.Agent.MCExtensionAgent import MCExtensionAgent as DIRACMCExtensionAgent
from LHCbDIRAC.ProductionManagementSystem.Client.ProductionRequestClient import ProductionRequestClient
from LHCbDIRAC.TransformationSystem.Client.TransformationClient import TransformationClient
from LHCbDIRAC.Workflow.Modules.ModulesUtilities import (
    getCPUNormalizationFactorAvg,
    getEventsToProduce,
    getProductionParameterValue,
)

AGENT_NAME = "Transformation/MCExtensionAgent"


class MCExtensionAgent(DIRACMCExtensionAgent):
    """MCExtensionAgent."""

    def __init__(self, *args, **kwargs):
        """c'tor."""
        DIRACMCExtensionAgent.__init__(self, *args, **kwargs)

        self.rpcProductionRequest = None
        self.transClient = None

        self.enableFlag = True

        # default values
        self.cpuE = 1
        self.cpuTimeAvg = 200000
        self.cpuNormalizationFactorAvg = 1.0

        # Artificial boost of the number of events requested to be created
        self.extensionFactorBoost = 20  # Meaning 20% more than what is calculated

    #############################################################################
    def initialize(self):
        """Logs some parameters and initializes the clients."""
        self.extensionFactorBoost = self.am_getOption("extensionFactorBoost", self.extensionFactorBoost)

        self.rpcProductionRequest = ProductionRequestClient()
        self.transClient = TransformationClient()

        self.log.info(f"Will consider the following transformation types: {str(self.transformationTypes)}")
        self.log.info(f"Will create a maximum of {self.maxIterationTasks} tasks per iteration")

        return S_OK()

    #############################################################################
    def execute(self):
        """The MCExtensionAgent execution method."""

        self.enableFlag = self.am_getOption("EnableFlag", "True")
        if not self.enableFlag == "True":
            self.log.info("MCExtensionAgent is disabled by configuration option EnableFlag")
            return S_OK("Disabled via CS flag")

        # done every cycle, as they may have changed
        self._getCPUParameters()

        # get the production requests in which we are interested
        productionRequests = self.rpcProductionRequest.getProductionRequestSummary("Active", "Simulation")
        if productionRequests["OK"]:
            productionRequests = productionRequests["Value"]
            self.log.info(f"Requests considered: {', '.join([str(prod) for prod in productionRequests])}")
        else:
            message = f"RPC call to ProductionRequest service failed : {productionRequests['Message']}"
            self.log.error(message)
            return S_ERROR(message)

        for productionRequestID, productionRequestSummary in productionRequests.items():
            ret = self._checkProductionRequest(productionRequestID, productionRequestSummary)
            if not ret["OK"]:
                return ret

        return S_OK()

    #############################################################################

    def _getCPUParameters(self):
        """Get the CPUTimeAvg and CPUNormalizationFactorAvg from config, or as a
        fail-over, there are some defaults."""

        op = Operations()
        self.cpuTimeAvg = op.getValue("Transformations/cpuTimeAvg", self.cpuTimeAvg)
        self.log.verbose("cpuTimeAvg = %d" % self.cpuTimeAvg)

        try:
            self.cpuNormalizationFactorAvg = getCPUNormalizationFactorAvg()
            self.log.verbose("cpuNormalizationFactorAvg = %d" % self.cpuNormalizationFactorAvg)
        except RuntimeError:
            self.log.info(
                "Could not get CPUNormalizationFactorAvg from config, defaulting to %d" % self.cpuNormalizationFactorAvg
            )

    #############################################################################
    def _checkProductionRequest(self, productionRequestID, productionRequestSummary):
        """Check if a production request need to be extended and do it if
        needed."""

        # check if enough events have been produced
        missingEvents = productionRequestSummary["reqTotal"] - productionRequestSummary["bkTotal"]
        self.log.info("Missing events for production request %d: %d" % (productionRequestID, missingEvents))
        if productionRequestSummary["bkTotal"] > 0 and missingEvents <= 0:
            message = "Enough events produced for production request %d" % productionRequestID
            self.log.verbose(message)
            return S_OK(message)

        # get the associated productions/transformations progress
        productionsProgress = self.rpcProductionRequest.getProductionProgressList(int(productionRequestID))
        if productionsProgress["OK"]:
            productionsProgress = productionsProgress["Value"]
        else:
            message = f"Failed to get productions progress : {productionsProgress['Message']}"
            self.log.error(message)
            return S_ERROR(message)
        productionsProgress = productionsProgress["Rows"]
        self.log.verbose("Progress for production request %d: %s" % (productionRequestID, str(productionsProgress)))

        # get the informations for the productions/transformations
        productions = []
        simulation = None
        for productionProgress in productionsProgress:
            productionID = productionProgress["ProductionID"]
            production = self.transClient.getTransformation(productionID)
            if not production["OK"]:
                message = "Failed to get informations on production %d : %s" % (productionID, production["Message"])
                self.log.error(message)
                return S_ERROR(message)
            production = production["Value"]
            productions.append(production)

            # determine which one is the simulation production
            if production["Type"] in self.transformationTypes:
                simulation = production
                simulationID = productionID
                for prodProgress in productionsProgress:
                    if prodProgress["ProductionID"] == simulationID:
                        simulationProgress = prodProgress
                        self.log.info(
                            "Progress for the simulation production %d of request %d: %s"
                            % (simulationID, productionRequestID, str(simulationProgress))
                        )

        if simulation is None:
            message = "Failed to get simulation production for request %d" % productionRequestID
            self.log.error(message)
            return S_ERROR(message)

        if simulation["Status"].lower() != "idle":
            # the simulation is still producing events
            message = "Simulation for production request %d is not Idle (%s)" % (
                productionRequestID,
                simulation["Status"],
            )
            self.log.verbose(message)
            return S_OK(message)

        # Checking how long ago this production became 'Idle'
        res = self.transClient.getTransformationLogging(simulationID)
        if not res["OK"]:
            return res
        lastLoggingEntry = res["Value"][-1]
        if ("idle" in lastLoggingEntry["Message"].lower()) and (
            (datetime.datetime.utcnow() - lastLoggingEntry["MessageDate"]).seconds < 900
        ):
            self.log.verbose("Prod %d is in 'Idle' for less than 15 minutes, waiting a bit" % simulationID)
            return S_OK("Prod %d is in 'Idle' for less than 15 minutes, waiting a bit" % simulationID)

        if simulationProgress["BkEvents"] < productionRequestSummary["reqTotal"]:
            # the number of events produced by the simulation is of the order of the number of events requested
            # -> there is probably no stripping production, no extension factor necessary
            return self._extendProduction(simulation, 1.0, missingEvents)
        else:
            # the number of events produced by the simulation is more than the number
            # of events requested, yet events are missing -> there is probably a
            # stripping production, an extension factor is needed to account for
            # stripped events some events may still be processed (eg. merged), so
            # wait that all the productions are idle
            if all(production["Status"].lower() == "idle" for production in productions):
                try:
                    extensionFactor = float(simulationProgress["BkEvents"]) / float(productionRequestSummary["bkTotal"])
                    return self._extendProduction(simulation, extensionFactor, missingEvents)
                except ZeroDivisionError:
                    return S_OK()
            else:
                return S_OK()

    #############################################################################
    def _extendProduction(self, production, extensionFactor, eventsNeeded):
        """Extends a production to produce eventsNeeded*extensionFactor more
        events."""
        productionID = production["TransformationID"]

        cpuEProd = getProductionParameterValue(production["Body"], "CPUe")
        if cpuEProd is None:
            self.log.warn("CPUe for transformation %d is not set, skipping for now" % productionID)
            return S_OK()
        cpuE = int(round(float(cpuEProd)))

        self.log.info(
            "Extending production %d, that is still missing %d events. \
                    Extension factor = %d, boost = %d"
            % (productionID, eventsNeeded, extensionFactor, self.extensionFactorBoost)
        )

        eventsToProduce = eventsNeeded * extensionFactor * (float(100 + self.extensionFactorBoost) / 100)
        max_e = getEventsToProduce(cpuE, self.cpuTimeAvg, self.cpuNormalizationFactorAvg)
        numberOfTasks = int(math.ceil(float(eventsToProduce) / float(max_e)))
        self.log.info("Extending production %d by %d tasks" % (productionID, numberOfTasks))

        # extend the transformation by the determined number of tasks
        res = self.transClient.extendTransformation(productionID, numberOfTasks)
        if not res["OK"]:
            message = "Failed to extend transformation %d : %s" % (productionID, res["Message"])
            self.log.error(message)
            return S_ERROR(message)
        else:
            message = "Successfully extended transformation %d by %d tasks" % (productionID, numberOfTasks)
            self.log.info(message)

            res = self.transClient.setTransformationParameter(productionID, "Status", "Active")
            if not res["OK"]:
                message = "Failed to set transformation %d to Active" % productionID
                self.log.error(message)
                return S_ERROR(message)

            return S_OK(message)



#!/usr/bin/env python
###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""Show replicas for a (set of) LFNs."""

from DIRAC.Core.Base.Script import Script


@Script()
def main():
    from LHCbDIRAC.DataManagementSystem.Client.DMScript import DMScript

    dmScript = DMScript()
    dmScript.registerFileSwitches()
    Script.registerSwitch("a", "All", "  Also show inactive replicas")
    Script.registerSwitch("", "DiskOnly", "  Show only disk replicas")
    Script.registerSwitch("", "PreferDisk", "  If disk replica, don't show tape replicas")
    Script.registerSwitch("", "ForJobs", "  Select only replicas that can be used for jobs")

    Script.setUsageMessage(
        "\n".join(
            [
                __doc__,
                "Usage:",
                f"  {Script.scriptName} [option|cfgfile] [<LFN>] [<LFN>...]",
            ]
        )
    )

    Script.parseCommandLine(ignoreErrors=False)

    from LHCbDIRAC.DataManagementSystem.Client.ScriptExecutors import executeLfnReplicas
    from DIRAC import exit

    exit(executeLfnReplicas(dmScript))


if __name__ == "__main__":
    main()



#!/usr/bin/env python
###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
import DIRAC
from DIRAC.Core.Base.Script import Script


@Script()
def main():
    from LHCbDIRAC.DataManagementSystem.Client.DMScript import DMScript

    dmScript = DMScript()
    dmScript.registerBKSwitches()
    dmScript.registerFileSwitches()

    fixIt = False
    Script.registerSwitch("", "FixIt", "Set replicas problematic if needed")
    Script.setUsageMessage(
        """
  Check the integrity of the state of the storages and information in the File Catalogs
  for a given file or a collection of files.

  Usage:
     %s <lfn | fileContainingLfns> <SE> <status>
  """
        % Script.scriptName
    )

    Script.parseCommandLine()

    for opt, val in Script.getUnprocessedSwitches():
        if opt == "FixIt":
            fixIt = True

    from DIRAC import gLogger

    gLogger.setLevel("INFO")
    from LHCbDIRAC.DataManagementSystem.Client.DataIntegrityClient import DataIntegrityClient

    for lfn in Script.getPositionalArgs():
        dmScript.setLFNsFromFile(lfn)
    lfns = dmScript.getOption("LFNs")
    if not lfns:
        print("No LFNs given...")
        Script.showHelp(exitCode=1)

    integrityClient = DataIntegrityClient()
    res = integrityClient.catalogFileToBK(lfns)
    if not res["OK"]:
        gLogger.error(res["Message"])
        DIRAC.exit(1)
    replicas = res["Value"]["CatalogReplicas"]
    metadata = res["Value"]["CatalogMetadata"]
    res = integrityClient.checkPhysicalFiles(replicas, metadata, fixIt=fixIt)
    if not res["OK"]:
        gLogger.error(res["Message"])
        DIRAC.exit(1)
    DIRAC.exit(0)


if __name__ == "__main__":
    main()



#!/usr/bin/env python
###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""Return the BK path for the directories of a (list of) files."""

from DIRAC.Core.Base.Script import Script


@Script()
def main():
    from LHCbDIRAC.DataManagementSystem.Client.DMScript import DMScript

    dmScript = DMScript()
    dmScript.registerFileSwitches()
    Script.registerSwitch("", "Full", "   Print out full BK dictionary (default: print out BK path)")
    Script.registerSwitch("", "GroupBy=", "   Return a list of files per <metadata item>")
    Script.registerSwitch("", "GroupByPath", "   Return a list of files per BK path")
    Script.registerSwitch("", "GroupByProduction", "   Return a list of files per production")
    Script.registerSwitch("", "Summary", "   Only give the number of files in each group (default: GroupByPath)")
    Script.registerSwitch("", "List", "   Print a list of group keys")
    Script.registerSwitch("", "IgnoreFileType", "   Ignore file type in path (useful for stripping)")
    Script.setUsageMessage(
        __doc__
        + "\n".join(
            [
                "Usage:",
                f"  {Script.scriptName} [option|cfgfile] ... LFN|File",
                "Arguments:",
                "  LFN:      Logical File Name",
                "  File:     Name of the file with a list of LFNs",
            ]
        )
    )
    Script.parseCommandLine()

    from LHCbDIRAC.BookkeepingSystem.Client.ScriptExecutors import executeFilePath

    executeFilePath(dmScript)


if __name__ == "__main__":
    main()



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""Minimal test for TransformationCleaningAgent."""
import importlib
import unittest

from unittest.mock import MagicMock
from DIRAC import S_OK, gLogger

from LHCbDIRAC.TransformationSystem.Agent.TransformationCleaningAgent import TransformationCleaningAgent


class TransformationCleaningAgentTestCase(unittest.TestCase):
    def setUp(self):
        self.mockAM = MagicMock()
        self.mockAM.am_getOption.return_value = S_OK()
        self.agent = importlib.import_module("LHCbDIRAC.TransformationSystem.Agent.TransformationCleaningAgent")
        self.agent.AgentModule = self.mockAM
        self.mockDTCAObj = MagicMock()
        self.mockDTCAObj.getTransformationDirectories.return_value = {
            "OK": True,
            "Value": ["/this/is/SIM/dir", "/this/is/FOO/dir", "/this/is/BAR/dir"],
        }
        self.mockDTCA = MagicMock()
        self.mockDTCA.return_value = self.mockDTCAObj
        self.agent.DiracTCAgent = self.mockDTCA
        self.agent = TransformationCleaningAgent("a", "b")
        self.agent.log = gLogger
        self.mockSUC = MagicMock()
        self.mockSUC.getStorageDirectories.return_value = {"OK": True, "Value": []}
        self.agent.storageUsageClient = self.mockSUC

    def test_getTransformationDirectories(self):
        res = self.agent.getTransformationDirectories(1)
        self.assertTrue(res["OK"])


if __name__ == "__main__":
    suite = unittest.defaultTestLoader.loadTestsFromTestCase(TransformationCleaningAgentTestCase)
    testResult = unittest.TextTestRunner(verbosity=2).run(suite)



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""DIRAC ProductionRequestDB class is a front-end to the repository database
containing Production Requests and other related tables."""
# Defined states:
# 'New'
# 'Rejected'
# 'Submitted'
# 'PPG OK'
# 'On-hold'
# 'Tech OK'
# 'Accepted'
# 'Active'
# 'Completed'
# 'Done'
# 'Cancelled'

import json
import time
import threading

from DIRAC import gLogger, S_OK, S_ERROR
from DIRAC.Core.Base.DB import DB

from LHCbDIRAC.ProductionManagementSystem.Utilities.Utils import informPeople


class ProductionRequestDB(DB):
    """DB class for ProductionManagement/ProductionRequestDB."""

    def __init__(self, parentLogger=None):
        """Constructor."""
        DB.__init__(self, "ProductionRequestDB", "ProductionManagement/ProductionRequestDB", parentLogger=parentLogger)
        self.dateColumns = ["StartingDate", "FinalizationDate"]
        self.dateFormat = "%Y-%m-%d"
        self.lock = threading.Lock()

    # Production Requests table
    requestFields = [
        "RequestID",
        "ParentID",
        "MasterID",
        "RequestAuthor",
        "RequestName",
        "RequestType",
        "RequestState",
        "RequestPriority",
        "RequestPDG",
        "RequestWG",
        "SimCondition",
        "SimCondID",
        "SimCondDetail",
        "ProPath",
        "ProID",
        "ProDetail",
        "EventType",
        "NumberOfEvents",
        "Description",
        "Comments",
        "Inform",
        "RealNumberOfEvents",
        "IsModel",
        "Extra",
        "RetentionRate",
        "FastSimulationType",
        "StartingDate",
        "FinalizationDate",
        "HasSubrequest",
        "bk",
        "bkSrTotal",
        "bkTotal",  # runtime
        "rqTotal",
        "crTime",
        "upTime",
    ]  # runtime

    historyFields = ["RequestID", "RequestState", "RequestUser", "TimeStamp"]

    # !!! current _escapeValues is buggy !!! None and not using connection...
    # _insert use it, so I can't...
    def _fixedEscapeValues(self, inValues, escape=None):
        """This method used to insert null value to the db, if the inserted value
        is None.

        :param list inValues: list of values
        :param int escape: the index of the value, which will be not escaped.
        """
        result = self._escapeValues(inValues)
        if not result["OK"]:
            return result
        outValues = result["Value"]
        for i, x in enumerate(outValues):
            if x == "None" or str(x) == "" or x == '"None"':
                if escape and i == escape:
                    continue
                outValues[i] = "NULL"
        return S_OK(outValues)

    @staticmethod
    def __prefixComments(update, old, user):
        """Add Log style prefix to the record like change."""
        if not update:
            return update
        if not old:
            old = ""
        if not update.startswith(old) and not update.endswith(old):
            return update
        prefix = f"Comment by {user} on {time.strftime('%b %d, %Y')}: "
        if update.startswith(old):
            if old.rstrip():
                old = old.rstrip() + "\n\n"
            return old + prefix + update[len(old) :].lstrip()
        return prefix + update.lstrip()

    def __getRequestInfo(self, iD, connection):
        """Retrive info fields from specified ID Used to get ParentID information.

        id must be checked before
        NOTE: it does self.lock.release() in case of errors
        """
        inFields = ["RequestState", "ParentID", "MasterID", "RequestAuthor", "Inform", "IsModel"]
        result = self._query(
            f"SELECT {','.join(inFields)} " + "FROM ProductionRequests " + f"WHERE RequestID={iD};", conn=connection
        )
        if not result["OK"]:
            self.lock.release()
            return result
        if len(result["Value"]) == 0:
            self.lock.release()
            return S_ERROR("Request does not exist")
        return S_OK(dict(zip(inFields, result["Value"][0])))

    def __getStateAndAuthor(self, iD, connection):
        """Return state, Author and inform list of Master for id (or id's own if no parents)
        id must be checked before
        NOTE: it does self.lock.release() in case of errors
        """
        result = self.__getRequestInfo(iD, connection)
        if not result["OK"]:
            return result
        pinfo = result["Value"]
        if not pinfo["MasterID"]:
            return S_OK([pinfo["RequestState"], pinfo["RequestAuthor"], pinfo["Inform"], pinfo["IsModel"]])
        result = self.__getRequestInfo(pinfo["MasterID"], connection)
        if not result["OK"]:
            return result
        pinfo = result["Value"]
        return S_OK([pinfo["RequestState"], pinfo["RequestAuthor"], pinfo["Inform"], pinfo["IsModel"]])

    def __checkMaster(self, master, iD, connection):
        """Return State of Master for id (or id's own if no parents) id and master
        must be checked before.

        It check that master can
        be reached with ParentID links.
        NOTE: it does self.lock.release() in case of errors
        """
        while True:
            result = self.__getRequestInfo(iD, connection)
            if not result["OK"]:
                return result
            pinfo = result["Value"]
            if iD == master:
                return S_OK(pinfo["RequestState"])
            if pinfo["MasterID"] != master:
                self.lock.release()
                return S_ERROR("Wrong MasterID for this ParentID")
            if not pinfo["ParentID"] or pinfo["ParentID"] == iD:
                self.lock.release()
                return S_ERROR("Parent tree is broken. Please contact expert")
            iD = pinfo["ParentID"]

    def createProductionRequest(self, requestDict, creds):
        """Create new Production Request
        TODO: Complete check of content
        """

        rec = dict.fromkeys(self.requestFields[1:-9], None)

        for x in requestDict:
            if x in rec and str(requestDict[x]) != "":
                rec[x] = requestDict[x]  # set only known not empty fields
        if rec["NumberOfEvents"]:  # Set RealNumberOfEvents if specified
            try:
                num = int(rec["NumberOfEvents"])
                if num > 0:
                    rec["RealNumberOfEvents"] = num
            except ValueError:
                pass
            except TypeError:
                gLogger.warn("RealNumberOfEvents not defined")
        if not rec["MasterID"]:
            rec["RequestPDG"] = ""
            if not rec["RequestState"]:
                rec["RequestState"] = "New"
        else:
            rec["RequestPDG"] = None
            rec["RequestState"] = None

        if rec["RequestState"]:
            if not rec["RequestState"] in ["New", "BK Check", "Submitted"]:
                return S_ERROR(f"The request can't be created in '{rec['RequestState']}' state")
            if rec["RequestState"] != "New":
                # !!! full information check must be here, but currently in the JS...
                # so we only check EventType consistency
                if not rec["EventType"]:
                    return S_ERROR("Please specify Event type/number or add subrequest(s)")
        if "Comments" in rec:
            rec["Comments"] = self.__prefixComments(rec["Comments"], "", creds["User"])
        rec["IsModel"] = 0

        recl = [rec[x] for x in self.requestFields[1:-9]]
        result = self._fixedEscapeValues(recl)
        if not result["OK"]:
            return result
        recls = result["Value"]
        if recl[-1] == "NULL":  # This is FastSimulationType, which can be None
            recls[-1] = '"None"'

        for dateValues in self.dateColumns:
            recls.append(
                "STR_TO_DATE('%s','%s')"
                % (requestDict.get(dateValues, time.strftime(self.dateFormat)), self.dateFormat)
            )

        self.lock.acquire()  # transaction begin ?? may be after connection ??
        result = self._getConnection()
        if not result["OK"]:
            self.lock.release()
            return S_ERROR("Failed to get connection to MySQL: " + result["Message"])
        connection = result["Value"]

        if rec["MasterID"]:  # have to check ParentID and MasterID consistency
            try:
                masterID = int(rec["MasterID"])
            except ValueError:
                self.lock.release()
                return S_ERROR("MasterID is not a number")
            if not rec["ParentID"]:
                self.lock.release()
                return S_ERROR("MasterID can not be without ParentID")
            try:
                parentID = int(rec["ParentID"])
            except ValueError:
                self.lock.release()
                return S_ERROR("ParentID is not a number")
            result = self.__checkMaster(masterID, parentID, connection)
            if not result["OK"]:
                return result
            result = self.__getStateAndAuthor(masterID, connection)
            if not result["OK"]:
                return result
            requestState, requestAuthor, _requestInform, isModel = result["Value"]
            if requestState != "New":
                self.lock.release()
                return S_ERROR("Requests can't be modified after submission")
            if requestAuthor != creds["User"] and not (isModel and creds["Group"] == "lhcb_tech"):
                self.lock.release()
                return S_ERROR("Only request author can add subrequests")
        elif rec["ParentID"]:
            try:
                parentID = int(rec["ParentID"])
            except ValueError:
                self.lock.release()
                return S_ERROR("ParentID is not a number")
            result = self.__getStateAndAuthor(parentID, connection)
            if not result["OK"]:
                return result

        req = "INSERT INTO ProductionRequests ( " + ",".join(self.requestFields[1:-7])
        req += f" ) VALUES ( {','.join(recls)} );"
        result = self._update(req, conn=connection)
        if not result["OK"]:
            self.lock.release()
            return result
        req = "SELECT LAST_INSERT_ID();"
        result = self._query(req, conn=connection)
        if not result["OK"]:
            self.lock.release()
            return result
        requestID = int(result["Value"][0][0])

        # Update history for masters. Errors are not reported back to the user.
        if rec["RequestState"]:
            result = self._update(
                "INSERT INTO RequestHistory ("
                + ",".join(self.historyFields[:-1])
                + f") VALUES ( {requestID},'{str(rec['RequestState'])}','{str(rec['RequestAuthor'])}')",
                conn=connection,
            )
            if not result["OK"]:
                gLogger.error(result["Message"])
        self.lock.release()
        if rec["RequestState"] in ["BK Check", "Submitted"]:
            rec["RequestID"] = requestID
            informPeople(rec, "", rec["RequestState"], creds["User"], rec["Inform"])
        return S_OK(requestID)

    @staticmethod
    def __addMonitoring(req, order):
        """Append monitoring columns.

        Somehow tricky SQL. Most probable need optimizations, but ok for
        now.
        """
        rQuery = "SELECT t.*,MIN(rh.TimeStamp) AS crTime,"
        rQuery += "           MAX(rh.TimeStamp) AS upTime "
        rQuery += "FROM "
        rQuery += "(SELECT t.*,CAST(COALESCE(SUM(sr.RealNumberOfEvents),0)+"
        rQuery += "                COALESCE(t.RealNumberOfEvents,0) AS SIGNED)"
        rQuery += "           AS rqTotal "
        rQuery += " FROM "
        rQuery += " (SELECT t.*,CAST(COALESCE(SUM(t.bkSrTotal),0)+"
        rQuery += "                  COALESCE(t.bk,0) AS SIGNED) AS bkTotal FROM "
        rQuery += "  (SELECT t.*,CAST(LEAST(COALESCE(SUM(pp.BkEvents),0),"
        rQuery += "                   COALESCE(SUM(sr.RealNumberOfEvents),0)) AS SIGNED)"
        rQuery += "              AS bkSrTotal FROM "
        rQuery += "   (SELECT t.*, CAST(SUM(pp.BkEvents) AS SIGNED)"
        rQuery += f"                AS bk FROM ({req}) as t "
        rQuery += "    LEFT JOIN ProductionProgress as pp ON (pp.RequestID=t.RequestID "
        rQuery += "    AND pp.Used=1) GROUP BY t.RequestID) as t "
        rQuery += "   LEFT JOIN ProductionRequests AS sr ON t.RequestID=sr.MasterID "
        rQuery += "   LEFT JOIN ProductionProgress AS pp ON (sr.RequestID=pp.RequestID "
        rQuery += "   AND pp.Used=1) GROUP BY t.RequestID) AS t"
        rQuery += "  GROUP BY t.RequestID) AS t"
        rQuery += " LEFT JOIN ProductionRequests as sr ON sr.MasterID=t.RequestID "
        rQuery += " GROUP BY t.RequestID) as t"
        rQuery += " LEFT JOIN RequestHistory as rh ON rh.RequestID=t.RequestID "
        rQuery += " GROUP BY t.RequestID"

        return rQuery + order

    def getProductionRequest(
        self, requestIDList, subrequestsFor=0, sortBy="", sortOrder="ASC", offset=0, limit=0, filterIn=None
    ):
        """Get the Production Request(s) details.

        If requestIDList is not empty, only productions from the list are
        returned. Otherwise master requests are returned (without
        subrequests) or all subrequests of 'subrequestsFor' (when
        specified). Parameters with explicit types are assumed checked by
        service.
        """
        if filterIn is None:
            filterIn = {}
        try:  # test parameters
            for x in requestIDList:
                y = int(x)
        except ValueError:
            return S_ERROR("Bad parameters (all request IDs must be numbers)")
        idFilter = False
        try:  # test filters
            sfilter = []
            for x in filterIn:
                if x not in self.requestFields[:-7]:
                    return S_ERROR("bad field in filterIn")
                val = str(filterIn[x])
                if val:
                    val = "( " + ",".join(['"' + y + '"' for y in val.split(",")]) + ") "
                    if x == "RequestID":
                        # Collect also masters masters
                        req = "SELECT RequestID,MasterID FROM ProductionRequests "
                        req += f"WHERE RequestID IN {val}"
                        result = self._query(req)
                        if not result["OK"]:
                            return result
                        val = []
                        for y in result["Value"]:
                            if str(y[1]) != "None":
                                val.append(str(y[1]))
                            else:
                                val.append(str(y[0]))
                        if not val:
                            return S_OK({"Rows": [], "Total": 0})
                        val = "( " + ",".join(val) + ") "
                        idFilter = True
                    sfilter.append(f" t.{x} IN {val} ")
            sfilter = " AND ".join(sfilter)
        except Exception as e:
            return S_ERROR("Bad filter content " + str(e))

        if sortBy:
            if sortBy not in self.requestFields[:-7]:
                return S_ERROR("sortBy field does not exist")
            if sortOrder != "ASC":
                sortOrder = "DESC"

        fields = ",".join(["t." + x for x in self.requestFields[:-7]])
        req = f"SELECT {fields} ,COUNT(sr.RequestID) AS HasSubrequest "
        req += "FROM ProductionRequests as t "
        req += "LEFT JOIN ProductionRequests AS sr ON t.RequestID=sr.ParentID "
        req += "WHERE "

        if requestIDList:
            idlist = ",".join([str(x) for x in requestIDList])
            where = f"t.RequestID IN ({idlist})"
        else:
            if subrequestsFor:
                where = f"t.ParentID={subrequestsFor}"
            else:
                where = sfilter
                if not idFilter:
                    if where:
                        where += " AND t.ParentID IS NULL"
                    else:
                        where = "t.ParentID IS NULL"
        req += where
        req += " GROUP BY t.RequestID"
        order = ""
        if sortBy:
            # order have to be applyed twice: before LIMIT and at the end
            order = f" ORDER BY {sortBy} {sortOrder}"
            req += order
        if limit and not subrequestsFor:
            req += f" LIMIT {offset},{limit}"
        result = self._query(self.__addMonitoring(req, order))
        if not result["OK"]:
            return result

        rows = [dict(zip(self.requestFields, row)) for row in result["Value"]]
        total = len(rows)
        if limit:
            result = self._query(f"SELECT COUNT(*) FROM ProductionRequests AS t WHERE {where}")
            if not result["OK"]:
                return result
            total = result["Value"][0][0]
        return S_OK({"Rows": rows, "Total": total})

    def __checkUpdate(self, update, old, creds, connection):
        """Check that update is possible.

        Return dict with values for _inform_people (with
        state=='' in  case notification is not required)
        NOTE: unlock in case of errors
        """
        requestID = old["RequestID"]
        result = self.__getStateAndAuthor(requestID, connection)
        if not result["OK"]:
            return result
        requestState, requestAuthor, requestInform, isModel = result["Value"]
        rec = old.copy()
        rec.update(update)
        inform = {
            "rec": rec,
            "state": "",
            "author": str(requestAuthor),
            "oldstate": requestState,
            "inform": requestInform,
        }

        hasSubreq = False
        if not old["MasterID"]:
            result = self._query(
                "SELECT RequestID " + "FROM ProductionRequests " + f"WHERE MasterID={requestID}", conn=connection
            )
            if not result["OK"]:
                self.lock.release()
                return result
            if result["Value"]:
                hasSubreq = True

        if creds["Group"] in ["diracAdmin", "lhcb_admin"]:
            return S_OK(inform)

        # Tech expert can (un)mark any master request as model. But only explicitly.
        if creds["Group"] == "lhcb_tech" and not old["MasterID"]:
            if len(update) == 1 and "IsModel" in update:
                if str(update["IsModel"]) == "1":
                    update["IsModel"] = 1
                else:
                    update["IsModel"] = 0
                return S_OK(inform)
        else:
            if "IsModel" in update:
                del update["IsModel"]

        if requestState in ["Done", "Cancelled"] and (creds["Group"] not in ["diracAdmin", "lhcb_admin"]):
            self.lock.release()
            return S_ERROR("Done or cancelled requests can't be modified")

        # Check that a person can update in general (that also means he can
        # change at least comments)
        if requestState in ["New", "BK OK", "Rejected"]:
            if requestAuthor != creds["User"] and not (isModel and creds["Group"] == "lhcb_tech"):
                self.lock.release()
                return S_ERROR("Only author is allowed to modify unsubmitted request")
        elif requestState == "BK Check":
            if creds["Group"] != "lhcb_bk":
                self.lock.release()
                return S_ERROR("Only BK expert can manage new Simulation Conditions")
        elif requestState == "Submitted":
            if creds["Group"] not in {"lhcb_ppg", "lhcb_tech", "lhcb_prmgr"}:
                self.lock.release()
                return S_ERROR("Only PPG members or Tech. experts are allowed to sign submitted request")
        elif requestState == "PPG OK":
            if creds["Group"] != "lhcb_tech":
                self.lock.release()
                return S_ERROR("Only Tech. experts are allowed to sign this request")
        elif requestState == "On-hold":
            if creds["Group"] != "lhcb_tech":
                self.lock.release()
                return S_ERROR("Only Tech. experts are allowed to sign this request")
        elif requestState == "Tech OK":
            if creds["Group"] != "lhcb_ppg":
                self.lock.release()
                return S_ERROR("Only PPG members are allowed to sign this request")
        elif requestState == "Accepted":
            if creds["Group"] != "lhcb_prmgr":
                self.lock.release()
                return S_ERROR("Only Tech. experts are allowed to manage accepted request")
        elif requestState in ["Active", "Completed"]:
            if not creds["Group"] in ["lhcb_prmgr", "lhcb_prod"]:
                self.lock.release()
                return S_ERROR("Only experts are allowed to manage active request")
        elif requestState in ["Done", "Cancelled"]:
            if creds["Group"] not in ["diracAdmin", "lhcb_admin"]:
                self.lock.release()
                return S_ERROR("Only admin can violate the system logic")
        else:
            self.lock.release()
            return S_ERROR(f"The request is in unknown state '{requestState}'")

        if old["MasterID"]:  # for subrequests it's simple
            if requestState == "New":
                for x in update:
                    if x not in ["EventType", "NumberOfEvents", "RealNumberOfEvents", "Comments"]:
                        self.lock.release()
                        return S_ERROR(f"{x} is not allowed in subrequests")
                    if x != "Comments" and not update[x]:
                        self.lock.release()
                        return S_ERROR("You must specify event type and number")
            else:
                if len(update) != 1 or "Comments" not in update:
                    self.lock.release()
                    return S_ERROR("Only comments can be changed for subrequest in progress")
            return S_OK(inform)
        # for masters it is more complicated...
        if requestState == "New":
            if "RequestState" not in update:
                return S_OK(inform)
            if update["RequestState"] == "Submitted":
                # !!! full information check must be here, but currently in the JS...
                # so we only check EventType consistency
                eventType = old["EventType"]
                if "EventType" in update:
                    eventType = update["EventType"]
                # gLogger.error(str(update))
                if eventType and hasSubreq:
                    self.lock.release()
                    return S_ERROR("The request has subrequests, so it must not specify Event type")
                if not eventType and not hasSubreq:
                    self.lock.release()
                    return S_ERROR("Please specify Event type/number or add subrequest(s)")
            else:
                self.lock.release()
                _msgTuple = (requestState, update["RequestState"])
                return S_ERROR("The request is '%s' now, moving to '%s' is not possible" % _msgTuple)
        elif requestState == "Rejected":
            self.lock.release()
            return S_ERROR("Rejected requests cannot be resurrected")
        elif requestState in {"BK Check", "BK OK"}:
            self.lock.release()
            raise NotImplementedError("This shouldn't happen any more")
        elif requestState == "Submitted":
            if not update["RequestState"] in {"Tech OK", "Rejected", "Accepted", "PPG OK", "Active"}:
                self.lock.release()
                _msgTuple = (requestState, update["RequestState"])
                return S_ERROR("The request is '%s' now, moving to '%s' is not possible" % _msgTuple)
        elif requestState in ["PPG OK", "On-hold"]:
            for x in update:
                if x not in ["RequestState", "Comments", "Inform", "ProPath", "ProID", "ProDetail", "Extra"]:
                    self.lock.release()
                    return S_ERROR(f"{x} can't be modified during Tech signing")
            if "RequestState" not in update:
                return S_OK(inform)
            if update["RequestState"] == "Tech OK":
                update["RequestState"] = "Accepted"
            if not update["RequestState"] in ["Accepted", "Rejected", "On-hold"]:
                self.lock.release()
                _msgTuple = (requestState, update["RequestState"])
                return S_ERROR("The request is '%s' now, moving to '%s' is not possible" % _msgTuple)
        # AZ: removed from logic
        #      if update['RequestState'] == 'Accepted' and not update.get('ProID',old['ProID']):
        #        self.lock.release()
        #        return S_ERROR("Registered processing pass is required to sign for Tech OK")
        elif requestState == "Tech OK":
            for x in update:
                if x not in ["RequestState", "RequestWG", "Comments", "Inform", "RequestPriority", "Extra"]:
                    self.lock.release()
                    return S_ERROR(f"{x} can't be modified during PPG signing")
            if "RequestState" not in update:
                return S_OK(inform)
            if update["RequestState"] == "PPG OK":
                update["RequestState"] = "Accepted"
            if not update["RequestState"] in ["Accepted", "Rejected"]:
                self.lock.release()
                _msgTuple = (requestState, update["RequestState"])
                return S_ERROR("The request is '%s' now, moving to '%s' is not possible" % _msgTuple)
        elif requestState in ["Accepted", "Active", "Completed"]:
            for x in update:
                if x not in ["RequestState", "Comments", "Inform", "ProPath", "ProID", "ProDetail", "Extra"]:
                    self.lock.release()
                    return S_ERROR(f"{x} can't be modified during the progress")
            if "RequestState" not in update:
                return S_OK(inform)
            if requestState == "Accepted":
                if not update["RequestState"] in ["Active", "Cancelled", "PPG OK"]:
                    self.lock.release()
                    _msgTuple = (requestState, update["RequestState"])
                    return S_ERROR("The request is '%s' now, moving to '%s' is not possible" % _msgTuple)
            elif requestState == "Active":
                if (update["RequestState"] not in ["Done", "Cancelled", "Completed", "Accepted"]) or (
                    update["RequestState"] != "Accepted" and creds["Group"] != "lhcb_prmgr"
                ):
                    self.lock.release()
                    _msgTuple = (requestState, update["RequestState"])
                    return S_ERROR("The request is '%s' now, moving to '%s' is not possible" % _msgTuple)
            else:
                if not update["RequestState"] in ["Active", "Done", "Cancelled"]:
                    self.lock.release()
                    _msgTuple = (requestState, update["RequestState"])
                    return S_ERROR("The request is '%s' now, moving to '%s' is not possible" % _msgTuple)
        inform["state"] = update["RequestState"]
        inform["rec"].update(update)
        return S_OK(inform)

    def updateProductionRequest(self, requestID, requestDict, creds):
        """Update existing production request In states other than New only state
        and comments are changable.

        TODO: RequestPDG change in ??? state
              Protect fields in subrequests
        """
        fdict = dict.fromkeys(self.requestFields[4:-7], None)
        rec = {}
        for x in requestDict:
            if x in fdict:
                if requestDict[x]:
                    rec[x] = requestDict[x]  # set only known fields
                else:
                    rec[x] = None  # to be more deterministic...

        self.lock.acquire()  # transaction begin ?? may be after connection ??
        result = self._getConnection()
        if not result["OK"]:
            self.lock.release()
            return S_ERROR("Failed to get connection to MySQL: " + result["Message"])
        connection = result["Value"]

        fields = ",".join(["t." + x for x in self.requestFields[:-7]])
        req = f"SELECT {fields} "
        req += "FROM ProductionRequests as t "
        req += f"WHERE t.RequestID={requestID}"
        result = self._query(req, conn=connection)
        if not result["OK"]:
            self.lock.release()
            return result
        if not result["Value"]:
            self.lock.release()
            return S_ERROR("The request is no longer exist")

        old = dict(zip(self.requestFields[:-7], result["Value"][0]))

        update = {}  # Decide what to update (and if that is required)
        for x in rec:
            if x in ("ProDetail", "SimCondDetail", "Extra"):
                if rec[x] == old[x]:
                    continue
                try:
                    recx = json.loads(rec[x])
                    oldx = json.loads(old[x])
                    if x == "SimCondDetail":
                        # Parameters without values are discarded by Web part,
                        # extra checks are for backward compatibility
                        if "inSMOG2State" in oldx and oldx["inSMOG2State"] is None and "inSMOG2State" not in recx:
                            recx["inSMOG2State"] = None
                        if "inExtendedDQOK" in oldx and oldx["inExtendedDQOK"] is None and "inExtendedDQOK" not in recx:
                            recx["inExtendedDQOK"] = None
                    if recx == oldx:
                        continue
                except TypeError:
                    # This happens if, for example, oldx is None (meaning there was not prodetail, while now there is).
                    # Which means that now we can update
                    pass
            elif str(rec[x]) == str(old[x]):
                continue

            if x == "RetentionRate" and float(rec[x]) == old[x]:
                continue

            update[x] = rec[x]

        if len(update) == 0:
            self.lock.release()
            return S_OK(requestID)  # nothing to update

        if "NumberOfEvents" in update:  # Update RealNumberOfEvents if specified
            num = 0
            try:
                num = int(rec["NumberOfEvents"])
                if num < 0:
                    num = 0
            except ValueError:
                pass
            except TypeError:
                gLogger.warn(f"NumberOfEvents not defined for {requestID}")
            update["RealNumberOfEvents"] = str(num)

        if "Comments" in update:
            update["Comments"] = self.__prefixComments(update["Comments"], old["Comments"], creds["User"])

        result = self.__checkUpdate(update, old, creds, connection)
        if not result["OK"]:
            return result
        inform = result["Value"]

        # we have to escape values... tricky way
        # in addition we can not escape the datetime.date values. so if they are being updated,
        # they will be added to this list later
        recl_fields = [column for column in update if column not in self.dateColumns]
        recl = [update[x] for x in recl_fields]
        result = self._fixedEscapeValues(recl)
        if not result["OK"]:
            self.lock.release()
            return result
        updateValues = result["Value"]
        for dateValue in self.dateColumns:
            # we are going to check if we have column which type is date
            # the order is very important that's why we use recl_fields
            if dateValue in self.dateColumns and requestDict.get(dateValue):
                recl_fields.append(dateValue)
                # my MySQL does not like time part in strict mode...
                dateOnly = str(requestDict.get(dateValue)).split(" ")[0]
                updateValues.append(f"STR_TO_DATE('{dateOnly}','{self.dateFormat}')")
        updates = ",".join([x + "=" + y for x, y in zip(recl_fields, updateValues)])

        req = "UPDATE ProductionRequests "
        req += f"SET {updates} "
        req += f"WHERE RequestID={requestID}"
        result = self._update(req, conn=connection)
        if not result["OK"]:
            self.lock.release()
            return result

        if "RequestState" in update:
            result = self._update(
                "INSERT INTO RequestHistory ("
                + ",".join(self.historyFields[:-1])
                + f") VALUES ( {requestID},'{str(update['RequestState'])}','{str(creds['User'])}')",
                conn=connection,
            )
            if not result["OK"]:
                gLogger.error(result["Message"])

        result = S_OK()
        gLogger.info(str(update))

        if result["OK"]:
            self.lock.release()

        informPeople(**inform)
        return S_OK(requestID)

    def __getSubrequestsList(self, iD, master, connection):
        """Return list of all subrequests for this request
        NOTE: it does self.lock.release() in case of errors
        """
        result = self._query(
            "SELECT RequestID " + "FROM ProductionRequests " + f"WHERE ParentID={iD} and MasterID={master}",
            conn=connection,
        )
        if not result["OK"]:
            self.lock.release()
            return result
        sr = []
        for x in result["Value"]:
            sr.append(x[0])
            result = self.__getSubrequestsList(x[0], master, connection)
            if not result["OK"]:
                self.lock.release()
                return result
            sr += result["Value"]
        return S_OK(sr)

    def deleteProductionRequest(self, requestID, creds):
        """Delete existing production.

        Subrequests are deleted. Substructure is moved up in the tree.
        Available is New and Rejected states only
        """
        try:
            requestID = int(requestID)
        except ValueError:
            return S_ERROR("RequestID is not a number")
        self.lock.acquire()  # transaction begin ?? may be after connection ??
        result = self._getConnection()
        if not result["OK"]:
            self.lock.release()
            return S_ERROR("Failed to get connection to MySQL: " + result["Message"])
        connection = result["Value"]

        result = self.__getStateAndAuthor(requestID, connection)
        if not result["OK"]:
            return result
        requestState, requestAuthor, _requestInform, isModel = result["Value"]
        if creds["Group"] not in ["diracAdmin", "lhcb_admin"]:
            if requestAuthor != creds["User"] and not (isModel and creds["Group"] == "lhcb_tech"):
                self.lock.release()
                gLogger.error(f"{creds['User']} can't remove {requestAuthor} request")
                return S_ERROR("Only author can remove a request")
            if requestState != "New" and requestState != "Rejected":
                self.lock.release()
                return S_ERROR("Can not remove request in processing")
        result = self.__getRequestInfo(requestID, connection)
        if not result["OK"]:
            return result
        pinfo = result["Value"]
        parentID = pinfo["ParentID"]
        masterID = pinfo["MasterID"]

        upperID = parentID
        if not upperID:
            upperID = "NULL"

        # delete subrequests
        req = ""
        if not masterID:  # this request is a master
            # delete history
            result = self._update("DELETE FROM RequestHistory " + f"WHERE RequestID={requestID}")
            if not result["OK"]:
                self.lock.release()
                return result
            # delete tracking
            result = self._update(
                "DELETE FROM ProductionProgress "
                + "WHERE RequestID IN "
                + "(SELECT RequestID FROM ProductionRequests "
                + f"WHERE RequestID={requestID} OR MasterID={requestID}"
                + ")",
                conn=connection,
            )
            if not result["OK"]:
                self.lock.release()
                return result
            req = "DELETE FROM ProductionRequests "
            req += f"WHERE MasterID={requestID}"
        else:  # this request is subrequest
            result = self.__getSubrequestsList(requestID, masterID, connection)
            if not result["OK"]:
                return result
            rlist = result["Value"]
            # delete tracking
            result = self._update(
                "DELETE FROM ProductionProgress "
                + f"WHERE RequestID IN ({','.join([str(x) for x in rlist] + [str(requestID)])})"
            )
            if not result["OK"]:
                self.lock.release()
                return result
            if len(rlist):
                req = "DELETE FROM ProductionRequests "
                req += f"WHERE RequestID in ({','.join([str(x) for x in rlist])})"
        if req:
            result = self._update(req, conn=connection)
            if not result["OK"]:
                self.lock.release()
                return result

        # move substructure
        req = f"UPDATE ProductionRequests SET ParentID={upperID} "
        req += f"WHERE ParentID={requestID}"
        result = self._update(req, conn=connection)
        if not result["OK"]:
            self.lock.release()
            return result

        # finally delete us
        req = "DELETE FROM ProductionRequests "
        req += f"WHERE RequestID={requestID}"
        result = self._update(req, conn=connection)

        self.lock.release()

        if not result["OK"]:
            return result
        return S_OK(requestID)

    def __getRequest(self, requestID, connection):
        """retrive complete request record.

        NOTE: unlock in case of errors
        """
        fields = ",".join(["t." + x for x in self.requestFields[:-7]])
        req = f"SELECT {fields} "
        req += "FROM ProductionRequests as t "
        req += f"WHERE t.RequestID={requestID}"
        result = self._query(req, conn=connection)
        if not result["OK"]:
            self.lock.release()
            return result
        if not result["Value"]:
            self.lock.release()
            return S_ERROR("The request is no longer exist")
        rec = dict(zip(self.requestFields[:-7], result["Value"][0]))
        return S_OK(rec)

    @staticmethod
    def __clearProcessingPass(rec):
        """clear processing pass section."""
        rec["ProID"] = None
        nd = {}
        rec["ProDetail"] = json.dumps(nd)

    def __duplicateDeep(self, requestID, masterID, parentID, creds, connection, clearpp):
        """recurcive duplication function.

        NOTE: unlock in case of errors
        """

        result = self.__getRequest(requestID, connection)
        if not result["OK"]:
            self.lock.release()
            return result
        rec = result["Value"]
        if clearpp:
            self.__clearProcessingPass(rec)
        rec["IsModel"] = 0

        if masterID and not rec["MasterID"]:
            return S_OK("")  # substructured request

        if rec["ParentID"]:
            rec["ParentID"] = parentID
        if rec["MasterID"]:
            rec["MasterID"] = masterID
        if rec["RequestAuthor"]:
            rec["RequestAuthor"] = creds["User"]
        if rec["RequestState"]:
            rec["RequestState"] = "New"

        # Clear RealNumberOfEvents if required
        try:
            num = 0
            num = int(rec["NumberOfEvents"])
            if num < 0:
                num = 0
        except ValueError:
            pass
        except TypeError:
            gLogger.warn(f"NumberOfEvents is not defined for {requestID}")
        rec["RealNumberOfEvents"] = str(num)

        recl = [rec[x] for x in self.requestFields[1:-7]]
        escapeIndex = self.requestFields[1:-7].index("FastSimulationType")
        result = self._fixedEscapeValues(recl, escapeIndex)
        if not result["OK"]:
            self.lock.release()
            return result
        recls = result["Value"]

        req = "INSERT INTO ProductionRequests ( " + ",".join(self.requestFields[1:-7])
        req += f" ) VALUES ( {','.join(recls)} );"
        result = self._update(req, conn=connection)
        if not result["OK"]:
            self.lock.release()
            return result
        req = "SELECT LAST_INSERT_ID();"
        result = self._query(req, conn=connection)
        if not result["OK"]:
            self.lock.release()
            return result
        newRequestID = int(result["Value"][0][0])

        # Update history for masters. Errors are not reported back to the user.
        if rec["RequestState"]:
            result = self._update(
                "INSERT INTO RequestHistory ("
                + ",".join(self.historyFields[:-1])
                + f") VALUES ( {newRequestID},'{str(rec['RequestState'])}','{str(rec['RequestAuthor'])}')",
                conn=connection,
            )
            if not result["OK"]:
                gLogger.error(result["Message"])

        # now for subrequests
        if not masterID:
            masterID = newRequestID
        parentID = newRequestID

        req = "SELECT RequestID "
        req += "FROM ProductionRequests as t "
        req += f"WHERE t.ParentID={requestID}"
        result = self._query(req, conn=connection)
        if not result["OK"]:
            self.lock.release()
            return result
        for chID in [row[0] for row in result["Value"]]:
            result = self.__duplicateDeep(chID, masterID, parentID, creds, connection, False)
            if not result["OK"]:
                return result

        return S_OK(int(newRequestID))

    def duplicateProductionRequest(self, requestID, creds, clearpp):
        """Duplicate production request with all it's subrequests (but without
        substructure).

        If that is subrequest, master must be in New state and user must be
        the author. If clearpp is set, all details in the Processing pass
        (of the master) are cleaned.
        """
        try:
            requestID = int(requestID)
        except ValueError:
            return S_ERROR("RequestID is not a number")
        self.lock.acquire()  # transaction begin ?? may be after connection ??
        result = self._getConnection()
        if not result["OK"]:
            self.lock.release()
            return S_ERROR("Failed to get connection to MySQL: " + result["Message"])
        connection = result["Value"]

        result = self.__getRequestInfo(requestID, connection)
        if not result["OK"]:
            return result
        pinfo = result["Value"]
        parentID = pinfo["ParentID"]
        masterID = pinfo["MasterID"]

        if masterID:
            clearpp = False
            result = self.__getStateAndAuthor(requestID, connection)
            if not result["OK"]:
                return result
            requestState, requestAuthor, _requestInform, isModel = result["Value"]
            if requestState != "New" or (
                requestAuthor != creds["User"] and not (isModel and creds["Group"] == "lhcb_tech")
            ):
                self.lock.release()
                return S_ERROR("Can not duplicate subrequest of request in progress")

        result = self.__duplicateDeep(requestID, masterID, parentID, creds, connection, clearpp)
        if result["OK"]:
            self.lock.release()
        return result

    @staticmethod
    def __checkAuthorizeSplit(requestState, creds):
        """Check that current user is allowed to split in specified state."""
        if creds["Group"] in ["diracAdmin", "lhcb_admin"]:
            return S_OK()
        if (requestState in ["Submitted", "PPG OK", "On-hold"]) and creds["Group"] == "lhcb_tech":
            return S_OK()
        if requestState in ["Accepted", "Active", "Completed"] and creds["Group"] == "lhcb_prmgr":
            return S_OK()
        return S_ERROR("You are not allowed to split the request")

    def __moveChildDeep(self, requestID, masterID, setParent, connection):
        """Update parent for this request if setParent is True and update master
        for this and all subrequests."""
        if not setParent:
            raise NotImplementedError("This is not supported and probably never be")
        req = "UPDATE ProductionRequests "
        req += f"SET ParentID={str(masterID)},MasterID={str(masterID)} "
        req += f"WHERE RequestID=ParentID={str(masterID)},MasterID={str(masterID)}"
        result = self._update(req, conn=connection)
        if not result["OK"]:
            return result
        req = "SELECT RequestID,MasterID "
        req += "FROM ProductionRequests as t "
        req += f"WHERE t.ParentID={requestID}"
        result = self._query(req, conn=connection)
        if not result["OK"]:
            return result
        for ch in result["Value"]:
            if ch[1]:
                ret = self.__moveChildDeep(ch[0], masterID, False, connection)
                if not ret["OK"]:
                    gLogger.error(f"_moveChildDeep: can not move to {str(masterID)}: {ret['Message']}")
                # !!! Failing that will leave both requests inconsistant !!!
                # But since we have only one subrequest level now, it will never happened
        return S_OK()

    def splitProductionRequest(self, requestID, splitlist, creds):
        """Fully duplicate master production request with its history and
        reassociate first level subrequests from splitlist (with there subrequest
        structure).

        Substructures can not be moved. Only experts in appropriate request
        state can request the split.
        """
        try:
            requestID = int(requestID)
        except ValueError:
            return S_ERROR("RequestID is not a number")
        if not splitlist:
            return S_ERROR("Split list is empty")
        isplitlist = []
        try:
            isplitlist = [int(x) for x in splitlist]
        except ValueError:
            return S_ERROR("RequestID in split list is not a number")

        self.lock.acquire()  # transaction begin ?? may be after connection ??
        result = self._getConnection()
        if not result["OK"]:
            self.lock.release()
            return S_ERROR("Failed to get connection to MySQL: " + result["Message"])
        connection = result["Value"]

        result = self.__getRequest(requestID, connection)
        if not result["OK"]:
            self.lock.release()
            return result
        rec = result["Value"]

        # Check that operation is valid
        if rec["MasterID"]:
            self.lock.release()
            return S_ERROR("Can not duplicate subrequest of request in progress")
        result = self.__checkAuthorizeSplit(rec["RequestState"], creds)
        if not result["OK"]:
            self.lock.release()
            return result
        req = "SELECT RequestID,MasterID "
        req += "FROM ProductionRequests as t "
        req += f"WHERE t.ParentID={requestID}"
        result = self._query(req, conn=connection)
        if not result["OK"]:
            self.lock.release()
            return result
        fisplitlist = []
        keeplist = []
        for ch in result["Value"]:
            if ch[0] in isplitlist and ch[1] == requestID:
                fisplitlist.append(int(ch[0]))
            elif ch[1]:
                keeplist.append(int(ch[0]))
        if len(isplitlist) != len(fisplitlist):
            self.lock.release()
            return S_ERROR("Requested for spliting subrequests are no longer exist")
        if not keeplist:
            self.lock.release()
            return S_ERROR("You have to keep at least one subrequest")

        # Now copy the master
        recl = [rec[x] for x in self.requestFields[1:-7]]
        result = self._fixedEscapeValues(recl, 24)
        if not result["OK"]:
            self.lock.release()
            return result
        recls = result["Value"]
        req = "INSERT INTO ProductionRequests ( " + ",".join(self.requestFields[1:-7])
        req += f" ) VALUES ( {','.join(recls)} );"
        result = self._update(req, conn=connection)
        if not result["OK"]:
            self.lock.release()
            return result
        req = "SELECT LAST_INSERT_ID();"
        result = self._query(req, conn=connection)
        if not result["OK"]:
            self.lock.release()
            return result
        newRequestID = int(result["Value"][0][0])

        # Move subrequests (!! Errors are not fatal !!)
        rsplitlist = []
        for x in fisplitlist:
            if self.__moveChildDeep(x, newRequestID, True, connection)["OK"]:
                rsplitlist.append(x)
        # If nothing could be moved, remove the master
        if not rsplitlist:
            req = "DELETE FROM ProductionRequests "
            req += f"WHERE RequestID={str(newRequestID)}"
            result = self._update(req, conn=connection)
            self.lock.release()
            return S_ERROR("Could not move subrequests")

        # Copy the history (failures are not fatal since
        # it is hard to revert the previous changes...)
        req = "SELECT " + ",".join(self.historyFields) + f" FROM RequestHistory WHERE RequestID={requestID} "
        req += "ORDER BY TimeStamp"
        result = self._query(req, conn=connection)
        if not result["OK"]:
            gLogger.error(f"SplitProductionRequest: can not get history for {str(requestID)}: {result['Message']}")
        else:
            for x in result["Value"]:
                x = list(x)
                x[0] = newRequestID
                ret = self._update(
                    "INSERT INTO RequestHistory ("
                    + ",".join(self.historyFields)
                    + ") VALUES ( %s,'%s','%s','%s')" % tuple([str(y) for y in x]),
                    conn=connection,
                )
            if not ret["OK"]:
                gLogger.error("SplitProductionRequest: add history fail: %s", ["Message"])

        self.lock.release()
        return S_OK(newRequestID)

    progressFields = ["ProductionID", "RequestID", "Used", "BkEvents"]

    def getProductionProgress(self, requestID):
        """return the list of associated productions requestID must be Long and
        already checked."""
        req = f"SELECT * FROM ProductionProgress WHERE RequestID={requestID}"
        result = self._query(req)
        if not result["OK"]:
            return result

        rows = [dict(zip(self.progressFields, row)) for row in result["Value"]]
        total = len(rows)
        return S_OK({"Rows": rows, "Total": total})

    def addProductionToRequest(self, pdict):
        """Associate production to request.

        Existence of request is checked first.
        TODO: check requestState
        """
        try:
            for x in self.progressFields:
                pdict[x] = int(pdict[x])
        except ValueError:
            return S_ERROR("Bad parameters")

        self.lock.acquire()  # transaction begin ?? may be after connection ??
        result = self._getConnection()
        if not result["OK"]:
            self.lock.release()
            return S_ERROR("Failed to get connection to MySQL: " + result["Message"])
        connection = result["Value"]

        result = self.__getStateAndAuthor(pdict["RequestID"], connection)
        if not result["OK"]:
            self.lock.release()
            return result
        # requestState, requestAuthor, requestInform = result['Value']

        req = "INSERT INTO ProductionProgress ( "
        req += ",".join(self.progressFields)
        req += " ) VALUES ( "
        req += ",".join([str(pdict[x]) for x in self.progressFields])
        req += " )"
        result = self._update(req, conn=connection)
        self.lock.release()
        if not result["OK"]:
            return result
        return S_OK(pdict["ProductionID"])

    def removeProductionFromRequest(self, productionID):
        """Deassociate production."""
        req = "DELETE FROM ProductionProgress "
        req += f"WHERE ProductionID={str(productionID)}"
        result = self._update(req)
        if not result["OK"]:
            return result
        return S_OK(productionID)

    def useProductionForRequest(self, productionID, used):
        """Deassociate production."""
        used = int(used)
        req = "UPDATE ProductionProgress "
        req += f"SET Used={str(used)} "
        req += f"WHERE ProductionID={str(productionID)}"
        result = self._update(req)
        if not result["OK"]:
            return result
        return S_OK(productionID)

    def getRequestHistory(self, requestID):
        """return the list of state changes for the requests requestID must be Long
        and already checked."""
        req = "SELECT " + ",".join(self.historyFields)
        req += f" FROM RequestHistory WHERE RequestID={requestID} "
        req += "ORDER BY TimeStamp"
        result = self._query(req)
        if not result["OK"]:
            return result

        rows = [dict(zip(self.historyFields, row)) for row in result["Value"]]
        total = len(rows)
        return S_OK({"Rows": rows, "Total": total})

    def getTrackedProductions(self):
        """return a list of all productions associated with requests in 'Active' or
        'Completed' state."""
        req1 = "SELECT RequestID FROM ProductionRequests WHERE RequestState in ('Active','Completed')"
        req2 = "SELECT RequestID FROM ProductionRequests WHERE RequestState in ('Active','Completed')"
        req2 += f" OR MasterID in ({req1})"
        req = "SELECT ProductionID FROM ProductionProgress WHERE RequestID "
        req += f"in ({req2})"

        result = self._query(req)
        if not result["OK"]:
            return result
        values = [row[0] for row in result["Value"]]
        return S_OK(values)

    def updateTrackedProductions(self, update):
        """update tracked productions."""
        # check parameters
        try:
            for x in update:
                x["ProductionID"] = int(x["ProductionID"])
                x["BkEvents"] = int(x["BkEvents"])
        except ValueError:
            return S_ERROR("Bad parameters")
        except TypeError:
            return S_ERROR("Parameters not defined")

        for x in update:
            result = self._update(
                "UPDATE ProductionProgress "
                + f"SET BkEvents={x['BkEvents']} "
                + f"WHERE ProductionID={x['ProductionID']}"
            )
            if not result["OK"]:
                gLogger.info(f"Problem in updating progress. Not fatal: {result['Message']}")
        return S_OK("")

    def __trackedInputSQL(self, fields):
        req = f"SELECT {fields} "
        req += "FROM ProductionRequests as t WHERE"
        req += ' t.RequestState in ("Active","Completed")'
        req += " AND NumberOfEvents<0 "
        return self._query(req)

    def getTrackedInput(self):
        """return a list of all requests with dynamic input in 'Active' or
        'Completed' states."""

        fields = ",".join(["t." + x for x in self.requestFields[:-7]])
        result = self.__trackedInputSQL(fields)
        if not result["OK"]:
            return result
        rec = []
        for x in result["Value"]:
            res = dict(zip(self.requestFields[:-7], x))
            if res["SimCondDetail"]:
                res.update(json.loads(res["SimCondDetail"]))
            else:
                continue
            del res["SimCondDetail"]
            try:
                num = int(res["RealNumberOfEvents"])
            except ValueError:
                num = 0
            except TypeError:
                num = 0
            if num > 0:
                continue
            rec.append(res)
        return S_OK(rec)

    def updateTrackedInput(self, update):
        """update real number of input events."""
        # check parameters
        try:
            for x in update:
                x["RequestID"] = int(x["RequestID"])
                x["RealNumberOfEvents"] = int(x["RealNumberOfEvents"])
        except ValueError:
            return S_ERROR("Bad parameters")
        except TypeError:
            return S_ERROR("Parameters not defined")
        result = self.__trackedInputSQL("RequestID")
        if not result["OK"]:
            return result
        allowed = dict.fromkeys([x[0] for x in result["Value"]], None)
        skiped = []
        for x in update:
            if not x["RequestID"] in allowed:
                skiped.append(x["RequestID"])
                continue
            req = "UPDATE ProductionRequests "
            req += f"SET RealNumberOfEvents={str(x['RealNumberOfEvents'])} "
            req += f"WHERE RequestID={str(x['RequestID'])}"
            result = self._update(req)
            if not result["OK"]:
                skiped.append(x["RequestID"])
        if skiped:
            return S_ERROR(f"updateTrackedInput has skiped requests {','.join(skiped)}")
        return S_OK("")

    def getProductionList(self, requestID):
        """return a list of all productions associated with the request or any its
        subrequest."""
        req1 = f"SELECT RequestID FROM ProductionRequests WHERE MasterID={requestID}"
        req = "SELECT ProductionID FROM ProductionProgress WHERE RequestID "
        req += f"in ({req1}) OR RequestID={requestID}"

        result = self._query(req)
        if not result["OK"]:
            return result

        gLogger.info(result["Value"])

        values = [row[0] for row in result["Value"]]
        return S_OK(values)

    def getAllProductionProgress(self):
        """return a dictionary containing for each requestID the active productions
        and the number of events."""
        req = "SELECT RequestID, ProductionID, Used, BkEvents FROM ProductionProgress;"
        res = self._query(req)
        if not res["OK"]:
            return res
        sRequestInfo = {}
        for sRequestID, prodID, used, events in res["Value"]:
            if sRequestID not in sRequestInfo:
                sRequestInfo[sRequestID] = {}
            sRequestInfo[sRequestID][prodID] = {"Used": used, "Events": events}
        return S_OK(sRequestInfo)

    def getFilterOptions(self):
        """Return the dictionary with possible values for filter."""
        opts = {}
        for key, value in [
            ("State", "RequestState"),
            ("Type", "RequestType"),
            ("Author", "RequestAuthor"),
            ("EType", "EventType"),
            ("WG", "RequestWG"),
        ]:
            req = f"SELECT DISTINCT {value} FROM ProductionRequests "
            req += f"WHERE {value} IS NOT NULL "
            req += f"ORDER BY {value}"
            res = self._query(req)
            if not res["OK"]:
                return res
            opts[key] = [row[0] for row in res["Value"]]
        return S_OK(opts)



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
# import mock
# if mock.__version__ < '1.0.1':
#   print 'Invalid Mock version %s !' % mock.__version__
#   import sys
#   sys.exit( 0 )
#
# import unittest
#
# from DIRAC import gLogger
#
# import LHCbDIRAC.ProductionManagementSystem.Agent.ProductionStatusAgent as sut
#
# class AgentTestCase( unittest.TestCase ):
#   """ Base class for the Agent test cases
#   """
#   def setUp( self ):
#
#     # reload SoftwareUnderTest to drop old patchers
#     reload( sut )
#
#     patcher = mock.patch( 'DIRAC.Core.Base.AgentModule.AgentModule', autospec = True )
#     pStarted = patcher.start()
#     class AgentMocked():
#       def __init__( self, *args, **kwargs ):
#         for k, v in pStarted.__dict__.items():
#           setattr( self, k, v )
#
#     sut.AgentModule = AgentMocked
#     sut.ProductionStatusAgent.__bases__ = ( AgentMocked, )
#
#     self.psa = sut.ProductionStatusAgent( '', '', '' )
#     self.psa.log = gLogger
#
#   def tearDown( self ):
#
#     # Stop patchers
#     mock.patch.stopall()
#
# #############################################################################
# # ProductionStatusAgent.py
# #############################################################################
#
# class ProductionStatusSuccess( AgentTestCase ):
#
#   def test__checkActiveToIdle( self ):
#     self.psa.transClient = mock.Mock()
#     self.psa._getTransformations = mock.Mock()
#     self.psa._getTransformationTaskStats = mock.Mock()
#     self.psa._getTransformationFilesStats = mock.Mock()
#     self.psa._updateProductionStatus = mock.Mock()

#     ###########################################################################
#     # no productions
#
#     updatedProductions = []
#     prods = []
#
#     self.psa._getTransformations.return_value = prods
#
#     self.psa._checkActiveToIdle( updatedProductions )
#     self.assertFalse( self.psa._updateProductionStatus.called )
#     self.psa._updateProductionStatus.reset_mock()

#     simulations and other type, idle and non idle

#     updatedProductions = []
#     prods = [1L, 2L, 3L]
#     prodsInfo = {1L : {'Value': {'Type' : 'MCSimulation'}},
#                  2L : {'Value': {'Type' : 'MCSimulation'}},
#                  3L : {'Value': {'Type' : 'MCSimulation'}},
#                  4L : {'Value': {'Type' : 'Merge'}},
#                  5L : {'Value': {'Type' : 'Merge'}},
#                  6L : {'Value': {'Type' : 'Merge'}}}
#     tasksStats = {1L : {'Created' : 100, 'Submitted' : 100, 'Done' : 90, 'Failed' : 10},  # idle
#                   2L : {'Created' : 100, 'Submitted' : 50, 'Done' : 50},  # not idle
#                   3L : {'Created' : 100, 'Submitted' : 100, 'Done' : 50, 'Failed' : 10, 'Running' : 40}}  # not idle
#     filesStats = {1L : {'Processed' : 100},  # idle
#                   2L : {'Processes' : 0, 'Unused' : 100},  # not idle
#                   3L : {'Processed' : 50, 'Unused' : 25, 'Assigned' : 25},  # not idle
#                   4L : {'Processed' : 100},  # idle
#                   5L : {'Processes' : 0, 'Unused' : 100},  # not idle
#                   6L : {'Processed' : 50, 'Unused' : 25, 'Assigned' : 25}}  # not idle
#
#     self.psa._getTransformations.return_value = prods
#     self.psa.transClient.getTransformation.side_effect = lambda prod : prodsInfo[prod]
#     self.psa._getTransformationTaskStats.side_effect = lambda prod : tasksStats[prod]
#     self.psa._getTransformationFilesStats.side_effect = lambda prod : filesStats[prod]

#     self.psa._checkActiveToIdle( updatedProductions )


#   def test__evaluateProgress( self ):
#     prodReqSummary = {
#                       8733: {'bkTotal': 1090805L, 'master': 0, 'reqTotal': 2000000L},
#                       8744: {'bkTotal': 2090805L, 'master': 0, 'reqTotal': 2000000L},
#                       9050: {'bkTotal': 1600993L, 'master': 9048, 'reqTotal': 500000L},
#                       9237: {'bkTotal': 0L, 'master': 9235, 'reqTotal': 500000L}
#                       }
#
#     progressSummary = {
#                         8733L: {20940L: {'Events': 33916336L, 'Used': 0},
#                                 20941L: {'Events': 1245467L, 'Used': 0},
#                                 20942L: {'Events': 1090805L, 'Used': 1}},
#                         8744L: {20140L: {'Events': 33916336L, 'Used': 0},
#                                 20141L: {'Events': 1245467L, 'Used': 0},
#                                 20142L: {'Events': 2090805L, 'Used': 1}},
#                         9050L: {21034L: {'Events': 1616993L, 'Used': 0},
#                                 21035L: {'Events': 1600993L, 'Used': 1}},
#                         9237L: {21080L: {'Events': 0L, 'Used': 0},
#                                 21081L: {'Events': 0L, 'Used': 1}}
#                       }
#
#     doneAndUsed, doneAndNotUsed, notDoneAndUsed, notDoneAndNotUsed = self.psa._evaluateProgress( prodReqSummary,
#                                                                                                  progressSummary )
#     self.assertEqual( doneAndUsed, {21035L:9050, 20142L:8744} )
#     self.assertEqual( doneAndNotUsed, { 21034L:9050, 20140L:8744, 20141L:8744} )
#     self.assertEqual( notDoneAndUsed, {20942L:8733} )
#     self.assertEqual( notDoneAndNotUsed, {20940L:8733, 20941L:8733} )

#   def test__reqsMap( self ):
#     prodReqSummary = {
#                       8733: {'bkTotal': 1090805L, 'master': 0, 'reqTotal': 2000000L},
#                       8744: {'bkTotal': 2090805L, 'master': 0, 'reqTotal': 2000000L},
#                       11: {'bkTotal': 2090805L, 'master': 12, 'reqTotal': 2000000L},
#                       13: {'bkTotal': 2090805L, 'master': 14, 'reqTotal': 2000000L},
#                       9050: {'bkTotal': 1600993L, 'master': 9048, 'reqTotal': 500000L},
#                       9051: {'bkTotal': 1600993L, 'master': 9048, 'reqTotal': 500000L},
#                       9237: {'bkTotal': 0L, 'master': 9235, 'reqTotal': 500000L},
#                       2: {'bkTotal': 600L, 'master': 1, 'reqTotal': 500L},
#                       4: {'bkTotal': 2600L, 'master': 3, 'reqTotal': 500L},
#                       5: {'bkTotal': 1600L, 'master': 3, 'reqTotal': 500L},
#                       6: {'bkTotal': 200L, 'master': 3, 'reqTotal': 500L},
#                       8: {'bkTotal': 2090805L, 'master': 0, 'reqTotal': 2000000L},
#                       }
#
#     progressSummary = {
#                         8733L: {20940L: {'Events': 33916336L, 'Used': 0},
#                                 20941L: {'Events': 1245467L, 'Used': 0},
#                                 20942L: {'Events': 1090805L, 'Used': 1}},
#                         8744L: {20140L: {'Events': 33916336L, 'Used': 0},
#                                 20141L: {'Events': 1245467L, 'Used': 0},
#                                 20142L: {'Events': 2090805L, 'Used': 1}},
#                         9050L: {21034L: {'Events': 1616993L, 'Used': 0},
#                                 21035L: {'Events': 1600993L, 'Used': 1}},
#                         9051L: {21036L: {'Events': 1616993L, 'Used': 0},
#                                 21037L: {'Events': 1600993L, 'Used': 1}},
#                         9237L: {21080L: {'Events': 0L, 'Used': 0},
#                                 21081L: {'Events': 0L, 'Used': 1}},
#                         2L: {1000L: {'Events': 600L, 'Used': 1}},
#                         4L: {1001L: {'Events': 5600L, 'Used': 0},
#                              1002L: {'Events': 5600L, 'Used': 0},
#                              1003L: {'Events': 2600L, 'Used': 1}},
#                         5L: {1011L: {'Events': 1600L, 'Used': 1}},
#                         6L: {1021L: {'Events': 200L, 'Used': 1}},
#                         8L: {1031L: {'Events': 2090805L, 'Used': 1}},
#                         11L: {1101L: {'Events': 2090805L, 'Used': 0},
#                               1102L: {'Events': 2090805L, 'Used': 1}},
#                         13L: {1301L: {'Events': 2090805L, 'Used': 0},
#                               1302L: {'Events': 2090805L, 'Used': 1}},
#                       }
#
#     res = self.psa._getReqsMap( prodReqSummary, progressSummary )
#     self.assertEqual( res, {8733:{8733:[20940L, 20941L, 20942L]},
#                             8744:{8744:[20140L, 20141L, 20142L]},
#                             9048:{9050:[21034L, 21035L], 9051:[21036L, 21037L]},
#                             9235:{9237:[21080L, 21081L]},
#                             1:{2:[1000L]},
#                             3:{4:[1001L, 1002L, 1003L], 5:[1011L], 6:[1021L]},
#                             8:{8:[1031L]},
#                             12: {11: [1101L, 1102L]},
#                             14: {13: [1301L, 1302L]}
#
#                             } )


#############################################################################
# Test Suite run
#############################################################################

# if __name__ == '__main__':
#   suite = unittest.defaultTestLoader.loadTestsFromTestCase( AgentTestCase )
#   suite.addTest( unittest.defaultTestLoader.loadTestsFromTestCase( ProductionStatusSuccess ) )
#   testResult = unittest.TextTestRunner( verbosity = 2 ).run( suite )

# EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""GangaDataFile is a utility to create a Data file, to be used by ganga.

Givn input files, it will create something like::

  from Gaudi.Configuration import *
  from GaudiConf import IOHelper
  IOHelper("ROOT").inputFiles([
      "LFN:foo",
      "LFN:bar"
  ], clear=True)

  FileCatalog().Catalogs = ["xmlcatalog_file:pool_xml_catalog.xml"]
"""
import os

from DIRAC import gLogger

from LHCbDIRAC.BookkeepingSystem.Client.LHCB_BKKDBClient import LHCB_BKKDBClient


class GangaDataFile:
    """Creates ganga data file."""

    def __init__(self, fileName="data.py", xmlcatalog_file="pool_xml_catalog.xml", log=None):
        """initialize."""
        if not log:
            self.log = gLogger.getSubLogger("GangaDataFile")
        else:
            self.log = log

        self.fileName = fileName
        self.xmlcatalog_file = xmlcatalog_file

        try:
            os.remove(self.fileName)
        except OSError:
            pass

        self.log.info(f"Creating Ganga data file {self.fileName} from scratch")

    ################################################################################

    def generateDataFile(self, lfns, persistency=None):
        """generate the data file."""
        if isinstance(lfns, str) and lfns:
            lfns = [lfns]
        elif not isinstance(lfns, list):
            self.log.error("Was expecting a list")
            raise TypeError("Expected List")
        if not len(lfns):
            self.log.warn("No file generated: was expecting a non-empty list")
            raise ValueError("list empty")

        try:
            persistency = persistency.upper()
            # Create a fake LFN->PFN dictionary to give the persistency
            fakePfns = dict.fromkeys(lfns, {"pfntype": persistency})
        except AttributeError:
            fakePfns = None

        script = LHCB_BKKDBClient().writeJobOptions(
            lfns, optionsFile=self.fileName, catalog=self.xmlcatalog_file, savePfn=fakePfns
        )
        self.log.info(f"Created Ganga data file {self.fileName}")

        return script



###############################################################################
# (c) Copyright 2023 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "COPYING".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
from urllib.parse import urlparse

import pytest


@pytest.fixture
def bkdb(request, monkeypatch):
    # These imports need to be protected else the tests can only be ran for
    # editable installs
    import LHCbDIRAC
    from LHCbDIRAC.BookkeepingSystem.DB.OracleBookkeepingDB import OracleBookkeepingDB

    oracle_url = request.config.getoption("--oracle-url")
    if oracle_url is None:
        pytest.skip("Requires an Oracle DB connection string to be passed with --oracle-url")
    parsed = urlparse(oracle_url)
    assert parsed.scheme == "oracle", "Only Oracle connections are supported"

    monkeypatch.setattr(
        LHCbDIRAC.BookkeepingSystem.DB.OracleBookkeepingDB,
        "getDatabaseSection",
        lambda a, b: f"/Systems/{a}/Production/Databases/{b}",
    )

    # By default, oracle uses out-of-band (OOB) communication which uses a
    # random port for sending additional data. When running the DB in a docker
    # container this makes it impossible to connect to the DB from the host as
    # we can't predict what the port will be to forward it (and it changes for
    # every connection attempt). An alternative would be to use `--net=host`
    # but `disable_oob=True` is cleaner.
    yield OracleBookkeepingDB(
        username=parsed.username,
        password=parsed.password,
        host=f"{parsed.hostname}:{parsed.port or 1521}{parsed.path}",
        disable_oob=True,
    )



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""Bookkeeping Reporting module (just prepare the files, do not send them
(which is done in the uploadOutput)"""

import os
import re
import shlex
import time

from DIRAC import gLogger, S_OK, S_ERROR, gConfig
from DIRAC.Core.Utilities.Subprocess import systemCall
from DIRAC.Resources.Catalog.PoolXMLFile import getGUID
from DIRAC.Workflow.Utilities.Utils import getStepCPUTimes

import LHCbDIRAC
from LHCbDIRAC.Resources.Catalog.PoolXMLFile import getOutputType
from LHCbDIRAC.Workflow.Modules.ModuleBase import ModuleBase
from LHCbDIRAC.Core.Utilities.ProductionData import constructProductionLFNs
from LHCbDIRAC.Core.Utilities.XMLSummaries import XMLSummary, XMLSummaryError
from LHCbDIRAC.Core.Utilities.BookkeepingJobInfo import BookkeepingJobInfo


class BookkeepingReport(ModuleBase):
    """BookkeepingReport class."""

    def __init__(self, bkClient=None, dm=None):
        """Usual c'tor."""

        self.log = gLogger.getSubLogger("BookkeepingReport")

        super().__init__(self.log, bkClientIn=bkClient, dm=dm)

        self.simDescription = "NoSimConditions"
        self.eventType = ""
        self.poolXMLCatName = ""
        self.stepInputData = []
        self.firstStepInput = ""
        self.jobType = ""
        self.stepOutputs = []
        self.histogram = False
        self.xf_o = None

        self.ldate = None
        self.ltime = None
        self.ldatestart = None
        self.ltimestart = None

    ################################################################################

    def execute(
        self,
        production_id=None,
        prod_job_id=None,
        wms_job_id=None,
        workflowStatus=None,
        stepStatus=None,
        wf_commons=None,
        step_commons=None,
        step_number=None,
        step_id=None,
        saveOnFile=True,
    ):
        """Usual executor."""
        try:
            super().execute(
                production_id,
                prod_job_id,
                wms_job_id,
                workflowStatus,
                stepStatus,
                wf_commons,
                step_commons,
                step_number,
                step_id,
            )

            if not self._checkWFAndStepStatus():
                return S_OK()

            bkLFNs, logFilePath = self._resolveInputVariables()

            doc = self.__makeBookkeepingXML(bkLFNs, logFilePath)

            if saveOnFile:
                bfilename = f"bookkeeping_{self.step_id}.xml"
                with open(bfilename, "wb") as bfile:
                    bfile.write(doc)
            else:
                print(doc)

            return S_OK()

        except Exception as e:  # pylint:disable=broad-except
            self.log.exception("Failure in BookkeepingReport execute module", lException=e)
            self.setApplicationStatus(repr(e))
            return S_ERROR(str(e))

        finally:
            super().finalize()

    ################################################################################
    # AUXILIAR FUNCTIONS
    ################################################################################

    def _resolveInputVariables(self):
        """By convention the module parameters are resolved here."""

        super()._resolveInputVariables()
        super()._resolveInputStep()

        self.stepOutputs, _sot, _hist = self._determineOutputs()

        # # VARS FROM WORKFLOW_COMMONS ##

        if "outputList" in self.workflow_commons:
            for outputItem in self.stepOutputs:
                if outputItem not in self.workflow_commons["outputList"]:
                    self.workflow_commons["outputList"].append(outputItem)
        else:
            self.workflow_commons["outputList"] = self.stepOutputs

        if (
            "BookkeepingLFNs" in self.workflow_commons
            and "LogFilePath" in self.workflow_commons
            and "ProductionOutputData" in self.workflow_commons
        ):
            logFilePath = self.workflow_commons["LogFilePath"]
            bkLFNs = self.workflow_commons["BookkeepingLFNs"]

            if not isinstance(bkLFNs, list):
                bkLFNs = [i.strip() for i in bkLFNs.split(";")]

        else:
            self.log.info("LogFilePath / BookkeepingLFNs parameters not found, creating on the fly")
            result = constructProductionLFNs(self.workflow_commons, self.bkClient)
            if not result["OK"]:
                self.log.error("Could not create production LFNs", result["Message"])
                raise ValueError("Could not create production LFNs")

            bkLFNs = result["Value"]["BookkeepingLFNs"]
            logFilePath = result["Value"]["LogFilePath"][0]

        self.ldate = time.strftime("%Y-%m-%d", time.gmtime(time.time()))
        self.ltime = time.strftime("%H:%M:%S", time.gmtime(time.time()))

        if "StartTime" in self.step_commons:
            startTime = self.step_commons["StartTime"]
            self.ldatestart = time.strftime("%Y-%m-%d", time.gmtime(startTime))
            self.ltimestart = time.strftime("%H:%M:%S", time.gmtime(startTime))

        try:
            self.xf_o = self.step_commons["XMLSummary_o"]
        except KeyError:
            self.log.warn("XML Summary object not found, will try to create it (again?)")
            try:
                xmlSummaryFile = self.step_commons["XMLSummary"]
            except KeyError:
                self.log.warn("XML Summary file name not found, will try to guess it")
                xmlSummaryFile = "summary{}_{}_{}_{}.xml".format(
                    self.cleanedApplicationName,
                    self.production_id,
                    self.prod_job_id,
                    self.step_number,
                )
                self.log.warn(f"Trying {xmlSummaryFile}")
                if xmlSummaryFile not in os.listdir("."):
                    self.log.warn(f"XML Summary file {xmlSummaryFile} not found, will try to guess a second time")
                    xmlSummaryFile = f"summary{self.cleanedApplicationName}_{self.step_id}.xml"
                    self.log.warn(f"Trying {xmlSummaryFile}")
                    if xmlSummaryFile not in os.listdir("."):
                        self.log.warn(
                            f"XML Summary file {xmlSummaryFile} not found, will try to guess a third and last time"
                        )
                        xmlSummaryFile = f"summary{self.cleanedApplicationName}_{self.step_number}.xml"
                        self.log.warn(f"Trying {xmlSummaryFile}")
            try:
                self.xf_o = XMLSummary(xmlSummaryFile)
            except XMLSummaryError as e:
                self.log.warn("No XML summary available", f"{repr(e)}")
                self.xf_o = None

        return bkLFNs, logFilePath

    ################################################################################
    ################################################################################

    def __makeBookkeepingXML(self, bkLFNs, logFilePath):
        """Bookkeeping xml looks like this::

        <Job ConfigName="" ConfigVersion="" Date="" Time="">
          <TypedParameter Name="" Type="" Value=""/>
          ...
          <InputFile Name=""/>
          ...
          <OutputFile Name="" TypeName="" TypeVersion="">
            <Parameter Name="" Value=""/>
            ...
            <Replica Location="" Name=""/>
            ....
          </OutputFile>
          ...
          <SimulationCondition>
            <Parameter Name="" Value=""/>
          </SimulationCondition>
        </Job>

        """
        job_info = BookkeepingJobInfo(
            ConfigName=self.workflow_commons.get("configName", self.applicationName),
            ConfigVersion=self.workflow_commons.get("configVersion", self.applicationVersion),
            Date=self.ldate,
            Time=self.ltime,
        )
        # Generate TypedParams
        self.__generateTypedParams(job_info)
        # Generate InputFiles
        self.__generateInputFiles(job_info, bkLFNs)
        # Generate OutputFiles
        self.__generateOutputFiles(job_info, bkLFNs, logFilePath)
        # Generate SimulationConditions
        if self.applicationName == "Gauss":
            job_info.simulation_condition = self.simDescription
        return job_info.to_xml()

    ################################################################################

    def __generateTypedParams(self, job_info):
        """Set fields in job_info.typed_parameters"""
        exectime, cputime = getStepCPUTimes(self.step_commons)
        job_info.typed_parameters.CPUTIME = str(cputime)
        job_info.typed_parameters.ExecTime = str(exectime)

        try:
            job_info.typed_parameters.WNMEMORY = str(self.xf_o.memory)
        except AttributeError:
            pass

        diracPower = gConfig.getValue("/LocalSite/CPUNormalizationFactor", "0")
        job_info.typed_parameters.WNCPUHS06 = diracPower
        mjfPower = gConfig.getValue("/LocalSite/CPUScalingFactor", "0")
        # Trick to know that the value is obtained from MJF: # from diracPower
        if mjfPower != diracPower:
            job_info.typed_parameters.WNMJFHS06 = str(mjfPower)
        job_info.typed_parameters.NumberOfProcessors = str(self.numberOfProcessors)
        job_info.typed_parameters.Production = str(self.production_id)
        job_info.typed_parameters.DiracJobId = str(self.jobID)
        job_info.typed_parameters.Name = str(self.step_id)
        job_info.typed_parameters.JobStart = f"{self.ldatestart} {self.ltimestart}"
        job_info.typed_parameters.JobEnd = f"{self.ldate} {self.ltime}"
        job_info.typed_parameters.Location = self.siteName
        job_info.typed_parameters.JobType = self.jobType

        job_info.typed_parameters.ProgramName = self.applicationName
        job_info.typed_parameters.ProgramVersion = self.applicationVersion

        job_info.typed_parameters.FirstEventNumber = str(1)

        job_info.typed_parameters.StatisticsRequested = str(self.numberOfEvents)

        job_info.typed_parameters.StepID = str(self.BKstepID)

        try:
            noOfEvents = self.xf_o.inputEventsTotal if self.xf_o.inputEventsTotal else self.xf_o.outputEventsTotal
        except AttributeError:
            # This happens iff the XML summary can't be created (e.g. for merging MDF files)
            res = self.bkClient.getFileMetadata(self.stepInputData)
            if not res["OK"]:
                raise AttributeError("Can't get the BKK file metadata")
            noOfEvents = sum(fileMeta["EventStat"] for fileMeta in res["Value"]["Successful"].values())
        job_info.typed_parameters.NumberOfEvents = str(noOfEvents)

    ################################################################################

    def __generateInputFiles(self, job_info, bkLFNs):
        self.log.debug("Adding InputData: bkLFNs", bkLFNs)
        self.log.debug("Adding InputData: self.stepInputData", self.stepInputData)
        for inputname in self.stepInputData or []:
            for bkLFN in bkLFNs:
                if os.path.basename(bkLFN).lower() == os.path.basename(inputname).lower():
                    # preserve the case
                    inputF = os.path.join(os.path.dirname(os.path.normpath(bkLFN)), os.path.basename(inputname))
                    job_info.input_files.append(inputF)
                    break
            else:
                # inputname is an LFN
                job_info.input_files.append(inputname)

    ################################################################################

    def __generateOutputFiles(self, job_info, bkLFNs, logFilePath):
        if self.eventType is not None:
            eventtype = self.eventType
        else:
            self.log.warn("BookkeepingReport: no eventType specified")
            eventtype = "Unknown"
        self.log.info("Event type =", eventtype)

        outputs = []
        bkTypeDict = {}
        for stepOutput in self.stepOutputs:
            if "outputDataName" in stepOutput:
                outputs.append(((stepOutput["outputDataName"]), (stepOutput["outputDataType"])))
            if "outputBKType" in stepOutput:
                bkTypeDict[stepOutput["outputDataName"]] = stepOutput["outputBKType"]
        outputs.append(((self.applicationLog), ("LOG")))
        self.log.info("Found outputs", outputs)

        if isinstance(logFilePath, list):
            logFilePath = logFilePath[0]

        for output, outputtype in outputs:
            self.log.info("Looking at output", f"{output} with type {outputtype}")
            typeName = outputtype.upper()
            typeVersion = "1"
            fileStats = "0"
            if output in bkTypeDict:
                typeVersion = getOutputType(output, self.stepInputData)[output]
                self.log.info("Setting POOL XML catalog type", f"for {output} to {typeVersion}")
                typeName = bkTypeDict[output].upper()
                self.log.info(
                    "Setting explicit BK type version",
                    f"for {output} to {typeVersion} and file type to {typeName}",
                )
                fileStats, output = self._getFileStatsFromXMLSummary(output, outputtype)

            if not os.path.exists(output):
                self.log.error("Output file does not exist", f"Output file name: {output}")
                continue
            # Output file size
            if "size" not in self.step_commons or output not in self.step_commons["size"]:
                try:
                    outputsize = str(os.path.getsize(output))
                except OSError:
                    outputsize = "0"
            else:
                outputsize = self.step_commons["size"][output]

            if "md5" not in self.step_commons or output not in self.step_commons["md5"]:
                comm = "md5sum " + str(output)
                resultTuple = systemCall(0, shlex.split(comm))
                status = resultTuple["Value"][0]
                out = resultTuple["Value"][1]
            else:
                status = 0
                out = self.step_commons["md5"][output]

            if status:
                self.log.info(f"Failed to get md5sum of {str(output)}")
                self.log.info(str(out))
                md5sum = "000000000000000000000000000000000000"
            else:
                md5sum = out.split()[0]

            if "guid" not in self.step_commons or output not in self.step_commons["guid"]:
                guidResult = getGUID(output)
                guid = ""
                if not guidResult["OK"]:
                    self.log.error("Could not find GUID", f"for {output} with message {guidResult['Message']}")
                elif guidResult["generated"]:
                    self.log.warn(
                        "PoolXMLFile generated GUID(s) for the following files ", ", ".join(guidResult["generated"])
                    )
                    guid = guidResult["Value"][output]
                else:
                    guid = guidResult["Value"][output]
                    self.log.info("Setting POOL XML catalog GUID", f"for {output} to {guid}")
            else:
                guid = self.step_commons["guid"][output]

            if not guid:
                self.log.error("No GUID found", f"for {output}")
                raise NameError("No GUID found")

            # find the constructed lfn
            if re.search(r"\.log$", output):
                lfn = os.path.join(logFilePath, self.applicationLog)
            else:
                for outputLFN in bkLFNs:
                    if os.path.basename(outputLFN) == output:
                        lfn = outputLFN
                        break
                else:
                    self.log.error("Could not find LFN", f"for {output}")
                    raise NameError("Could not find LFN of output file")

            oldTypeName = None
            if "HIST" in typeName.upper():
                typeVersion = "0"

            # PROTECTION for old production XMLs
            # TODO: think about removing this!
            if typeName.upper() == "HIST":
                typeName = f"{self.applicationName.upper()}HIST"

            # Add Output to the XML file
            outputFile = BookkeepingJobInfo.OutputFile(
                Name=lfn,
                TypeName=typeName,
                TypeVersion=typeVersion,
                FileSize=outputsize,
                CreationDate=time.strftime("%Y-%m-%d %H:%M:%S", time.gmtime(time.time())),
                MD5Sum=md5sum,
                Guid=guid,
            )

            # HIST is in the dataTypes e.g. we may have new names in the future ;)
            if oldTypeName:
                typeName = oldTypeName

            if outputtype != "LOG":
                outputFile.EventTypeId = eventtype
                if fileStats != "Unknown":
                    outputFile.EventStat = fileStats

            # Log file replica information
            if self.applicationLog:
                logfile = self.applicationLog
                if logfile == output:
                    outputFile.Replica = f"http://lhcb-logs.cern.ch/storage{logFilePath}/{self.applicationLog}"

            job_info.output_files.append(outputFile)

    ################################################################################

    def _getFileStatsFromXMLSummary(self, output, outputtype):
        """Gets stats per file from the XML summary, considering files registered
        with different cases.

        :params str output: file name looked out
        :params str outputType: file type looked out
        :returns: (str, str) with stats and actual file name
        """
        try:
            return str(self.xf_o.outputsEvents[output]), output
        except AttributeError as e:
            # This happens iff the XML summary can't be created (e.g. for merging MDF files)
            self.log.warn(
                "XML summary not created, unable to determine the output events and setting to 'Unknown'", repr(e)
            )
            return "Unknown", output
        except KeyError as e:
            self.log.warn("Could not find output LFN in XML summary object", repr(e))

            # here starting to look if by chance the file has been produced with a different case
            for outputFileInXML in self.xf_o.outputsEvents:
                if output.lower() == outputFileInXML.lower():
                    self.log.info(
                        "Found output LFN in XML summary object with different case",
                        f"{output} -> {outputFileInXML}",
                    )
                    return str(self.xf_o.outputsEvents[outputFileInXML]), outputFileInXML

            if ("hist" in outputtype.lower()) or (".root" in outputtype.lower()):
                self.log.warn(
                    "HIST file not found in XML summary, event stats set to 'Unknown'", f"HIST not found = {output}"
                )
                return "Unknown", output

            raise KeyError("Could not find output LFN in XML summary object")



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""LHCb Job Class.

The LHCb Job class inherits generic VO functionality from the Job base class
and provides VO-specific functionality to aid in the construction of
workflows.

Helper functions are documented with example usage for the DIRAC API.

Below are several examples of LHCbJob usage.

An example DaVinci application script would be::

  from LHCbDIRAC.Interfaces.API.DiracLHCb import DiracLHCb
  from LHCbDIRAC.Interfaces.API.LHCbJob import LHCbJob

  j = LHCbJob()
  j.setCPUTime(5000)
  j.setApplication("DaVinci", "v19r12", "DaVinciv19r12.opts",
  optionsLine="ApplicationMgr.EvtMax=1",
  inputData=["/lhcb/production/DC06/phys-v2-lumi2/00001650/DST/0000/00001650_00000054_5.dst"])
  j.setName("MyJobName")
  j.setDestination("LCG.CERN.cern")

  dirac = DiracLHCb()
  jobID = dirac.submitJob(j)
  print("Submission Result:", jobID)

The setDestination() method is optional and takes the DIRAC site name as an argument.
"""
import os
import re

from DIRAC import S_OK, S_ERROR
from DIRAC.Interfaces.API.Job import Job
from DIRAC.ConfigurationSystem.Client.Helpers.Operations import Operations
from DIRAC.Workflow.Utilities.Utils import getStepDefinition, addStepToWorkflow

from LHCbDIRAC.BookkeepingSystem.Client.BookkeepingClient import BookkeepingClient
from LHCbDIRAC.Interfaces.API.DiracLHCb import DiracLHCb
from LHCbDIRAC.ConfigurationSystem.Client.Helpers.Resources import getPlatformForJob


class LHCbJob(Job):
    """LHCbJob class as extension of DIRAC Job class."""

    #############################################################################

    def __init__(self, script=None, stdout="std.out", stderr="std.err"):
        """Instantiates the Workflow object and some default parameters."""
        super().__init__(script, stdout, stderr)
        self.stepCount = 0
        self.inputDataType = "DATA"  # Default, other options are MDF, ETC
        self.importLocation = "LHCbDIRAC.Workflow.Modules"
        self.opsHelper = Operations()

    #############################################################################

    def setApplication(
        self,
        appName,
        appVersion,
        optionsFiles,
        inputData=None,
        optionsLine=None,
        inputDataType=None,
        logFile=None,
        events=-1,
        extraPackages=None,
        systemConfig="ANY",
        multicore=False,
        modulesNameList=None,
        parametersList=None,
        eventTimeout=None,
    ):
        """Specifies gaudi application for DIRAC workflows, executed via gaudirun.

        For LHCb these could be e.g. Gauss, Boole, Brunel, DaVinci, LHCb, etc.

        The optionsFiles parameter can be the path to an options file or a list of paths to options files.
        All options files are automatically appended to the job input sandBox but the first in the case of a
        list is assumed to be the 'master' options file.

        Input data for application script must be specified here, please note that if this is set at the job level,
        e.g. via setInputData() but not above, input data is not in the scope of the specified application.

        Any input data specified at the step level that is not already specified at the job level is added automatically
        as a requirement for the job.

        Example usage:

        >>> job = LHCbJob()
        >>> job.setApplication('DaVinci','v19r5',optionsFiles='MyDV.opts',
        inputData=['/lhcb/production/DC06/phys-lumi2/00001501/DST/0000/00001501_00000320_5.dst'],logFile='dv.log')

        :param appName: Application name
        :type appName: string
        :param appVersion: Application version
        :type appVersion: string
        :param optionsFiles: Path to options file(s) for application
        :type optionsFiles: string or list
        :param inputData: Input data for application (if a subset of the overall input data for a given job is required)
        :type inputData: single LFN or list of LFNs
        :param optionsLine: Additional options lines for application
        :type optionsLine: string
        :param inputDataType: Input data type for application (e.g. DATA, MDF, ETC)
        :type inputDataType: string
        :param logFile: Optional log file name
        :type logFile: string
        :param events: Optional number of events
        :type events: integer
        :param extraPackages: Optional extra packages
        :type extraPackages: string
        :param systemConfig: Optional CMT config
        :type systemConfig: string
        :param multicore: multicore flag: set True if the step CAN run in multicore mode
                                         (if set, and if possible, it will use the --ncups flag of gaudirun)
        :type extraPackages: bool
        :param modulesNameList: list of module names. The default is given.
        :type modulesNameList: list
        :param parametersList: list of parameters. The default is given.
        :type parametersList: list
        :param eventTimeout: event timeout, in seconds
        :type eventTimeout: int
        """
        kwargs = {
            "appName": appName,
            "appVersion": appVersion,
            "optionsFiles": optionsFiles,
            "inputData": inputData,
            "optionsLine": optionsLine,
            "inputDataType": inputDataType,
            "logFile": logFile,
        }
        if not isinstance(appName, str) or not isinstance(appVersion, str):
            return self._reportError("Expected strings for application name and version", __name__, **kwargs)

        if logFile:
            if isinstance(logFile, str):
                logName = logFile
            else:
                return self._reportError("Expected string for log file name", __name__, **kwargs)
        else:
            logName = f"{appName}_{appVersion}.log"

        if not inputDataType:
            inputDataType = self.inputDataType
        if not isinstance(inputDataType, str):
            return self._reportError("Expected string for input data type", __name__, **kwargs)

        optionsFile = None
        if not optionsFiles:
            return self._reportError("Expected string or list for optionsFiles", __name__, **kwargs)
        if isinstance(optionsFiles, str):
            optionsFiles = [optionsFiles]
        if not isinstance(optionsFiles, list):
            return self._reportError("Expected string or list for optionsFiles", __name__, **kwargs)
        for optsFile in optionsFiles:
            if not optionsFile:
                self.log.verbose(f"Found master options file {optsFile}")
                optionsFile = optsFile
            if os.path.exists(optsFile):
                self.log.verbose(f"Found specified options file: {optsFile}")
                self.addToInputSandbox.append(optsFile)
                optionsFile += f";{optsFile}"
            elif re.search(r"\$", optsFile):
                self.log.verbose(
                    f"Assuming {optsFile} is using an environment variable to be resolved during execution"
                )
                if not optionsFile == optsFile:
                    optionsFile += f";{optsFile}"
            else:
                return self._reportError(f"Specified options file {optsFile} does not exist", __name__, **kwargs)

        # ensure optionsFile list is unique:
        tmpList = optionsFile.split(";")
        optionsFile = ";".join(list(set(tmpList)))
        self.log.verbose(f"Final options list is: {optionsFile}")

        if inputData:
            if isinstance(inputData, str):
                inputData = [inputData]
            if not isinstance(inputData, list):
                return self._reportError(
                    "Expected single LFN string or list of LFN(s) for inputData", __name__, **kwargs
                )
            if inputData != ["previousStep"]:
                for i in range(len(inputData)):
                    inputData[i] = inputData[i].replace("LFN:", "")
                inputData = ["LFN:" + x for x in inputData]
                inputDataStr = ";".join(inputData)
                self.addToInputData.append(inputDataStr)

        self.stepCount += 1
        stepName = f"{appName}Step{self.stepCount}"

        if not modulesNameList:
            modulesNameList = [
                "CreateDataFile",
                "GaudiApplication",
                "FileUsage",
                "AnalyseFileAccess",
                "UserJobFinalization",
            ]
        if not parametersList:
            parametersList = [
                ("applicationName", "string", "", "Application Name"),
                ("applicationVersion", "string", "", "Application Version"),
                ("applicationLog", "string", "", "Name of the output file of the application"),
                ("optionsFile", "string", "", "Options File"),
                ("extraOptionsLine", "string", "", "This is appended to standard options"),
                ("inputDataType", "string", "", "Input Data Type"),
                ("inputData", "string", "", "Input Data"),
                ("numberOfEvents", "string", "", "Events treated"),
                ("multiCore", "string", "", "If the step can run multicore"),
                ("extraPackages", "string", "", "ExtraPackages"),
                ("SystemConfig", "string", "", "binary tag"),
                ("eventTimeout", "string", "", "event timeout"),
            ]
            if appName == "Gauss":
                # add parameter for gracefully stop Gauss with a signal (interpreted by the DIRAC Watchdog)
                parametersList.append(("StopSigNumber", "string", "", "signal number (interpreted by Gaudi)"))
                parametersList.append(("StopSigRegex", "string", "", "RegEx of what to stop"))

        step = getStepDefinition(
            stepName,
            modulesNameList=modulesNameList,
            importLine="LHCbDIRAC.Workflow.Modules",
            parametersList=parametersList,
        )

        logPrefix = f"Step{self.stepCount}_"
        logName = f"{logPrefix}{logName}"
        self.addToOutputSandbox.append(logName)

        stepInstance = addStepToWorkflow(self.workflow, step, stepName)

        stepInstance.setValue("applicationName", appName)
        stepInstance.setValue("applicationVersion", appVersion)
        stepInstance.setValue("applicationLog", logName)
        if optionsFile:
            stepInstance.setValue("optionsFile", optionsFile)
        if optionsLine:
            stepInstance.setValue("extraOptionsLine", optionsLine)
        if inputDataType:
            stepInstance.setValue("inputDataType", inputDataType)
        if inputData:
            stepInstance.setValue("inputData", ";".join(inputData))
        stepInstance.setValue("numberOfEvents", str(events))
        stepInstance.setValue("multiCore", "Y" if multicore else "N")
        stepInstance.setValue("extraPackages", extraPackages)
        stepInstance.setValue("SystemConfig", systemConfig)
        if eventTimeout:
            stepInstance.setValue("eventTimeout", eventTimeout)
        if appName == "Gauss":
            stepInstance.setValue("StopSigNumber", "2")
            stepInstance.setValue("StopSigRegex", ".* Gauss.* gaudirun.py .* prodConf_Gauss_[0-9]*_[0-9]*_[0-9].py")

        return S_OK(stepInstance)

    #############################################################################

    def setAncestorDepth(self, depth):
        """Helper function.

        Level at which ancestor files are retrieved from the bookkeeping.

        For analysis jobs running over RDSTs the ancestor depth may be specified
        to ensure that the parent DIGI / DST files are staged before job execution.

        Example usage:

        >>> job = LHCbJob()
        >>> job.setAncestorDepth(2)

        :param depth: Ancestor depth
        :type depth: string or int
        """
        kwargs = {"depth": depth}
        description = "Level at which ancestor files are retrieved from the bookkeeping"
        if isinstance(depth, str):
            try:
                self._addParameter(self.workflow, "AncestorDepth", "JDL", int(depth), description)
            except BaseException:
                return self._reportError("Expected integer for Ancestor Depth", __name__, **kwargs)
        elif isinstance(depth, int):
            self._addParameter(self.workflow, "AncestorDepth", "JDL", depth, description)
        else:
            return self._reportError("Expected integer for Ancestor Depth", __name__, **kwargs)
        self._addParameter(
            self.workflow,
            "JobPath",
            "JDL",
            "JobPath,JobSanity,InputData,AncestorFiles,JobScheduling",
            "Custom list of optimizers",
        )
        return S_OK()

    #############################################################################

    def setInputDataType(self, inputDataType):
        """Explicitly set the input data type to be conveyed to Gaudi Applications.

        Default is DATA, e.g. for DST / RDST files.  Other options include:
         - MDF, for .raw files
         - ETC, for running on a public or private Event Tag Collections.

        Example usage:

        >>> job = LHCbJob()
        >>> job.setInputDataType('ETC')

        :param inputDataType: Input Data Type
        :type inputDataType: String
        """
        description = "User specified input data type"
        if not isinstance(inputDataType, str):
            try:
                inputDataType = str(inputDataType)
            except TypeError:
                return self._reportError(
                    "Expected string for input data type", __name__, **{"inputDataType": inputDataType}
                )

        self.inputDataType = inputDataType
        self._addParameter(self.workflow, "InputDataType", "JDL", inputDataType, description)
        return S_OK()

    #############################################################################

    def setCondDBTags(self, condDict):
        """Under development. Helper function.

        Specify Conditions Database tags by by Logical File Name (LFN).

        The input dictionary is of the form: {<DB>:<TAG>} as in the example below.

        Example usage:

        >>> job = LHCbJob()
        >>> job.setCondDBTags({'DDDB':'DC06','LHCBCOND':'DC06'})

        :param condDict: CondDB tags
        :type condDict: Dict of DB, tag pairs
        """
        kwargs = {"condDict": condDict}
        if not isinstance(condDict, dict):
            return self._reportError("Expected dictionary for CondDB tags", __name__, **kwargs)

        conditions = []
        for db, tag in condDict.items():
            try:
                db = str(db)
                tag = str(tag)
                conditions.append(".".join([db, tag]))
            except BaseException:
                return self._reportError("Expected string for conditions", __name__, **kwargs)

        condStr = ";".join(conditions)
        description = "List of CondDB tags"
        self._addParameter(self.workflow, "CondDBTags", "JDL", condStr, description)
        return S_OK()

    #############################################################################

    def setOutputData(self, lfns, OutputSE=None, OutputPath=None, replicate=None, filePrepend=None):
        """Helper function, used in preference to Job.setOutputData() for LHCb.

        For specifying user output data to be registered in Grid storage.

        Example usage:

        .. code-block:: python

          >>> job = Job()
          >>> job.setOutputData(['DVNtuple.root'])

        :param lfns: Output data file or files
        :type lfns: Single string or list of strings ['','']
        :param OutputSE: Optional parameter to specify the Storage
        :param OutputPath: Optional parameter to specify the Path in the Storage
          Element to store data or files, e.g. CERN-tape
        :type OutputSE: string or list
        :type OutputPath: string
        """
        # FIXME: the output data as specified here will be treated by the UserJobFinalization module
        # If we remove this method (which is totally similar to the Job() one, the output data will be
        # treated by the JobWrapper. So, can and maybe should be done, but have to pay attention
        kwargs = {"lfns": lfns, "OutputSE": OutputSE, "OutputPath": OutputPath}
        if isinstance(lfns, list) and lfns:
            outputDataStr = ";".join(lfns)
            description = "List of output data files"
            self._addParameter(self.workflow, "UserOutputData", "JDL", outputDataStr, description)
        elif isinstance(lfns, str):
            description = "Output data file"
            self._addParameter(self.workflow, "UserOutputData", "JDL", lfns, description)
        else:
            return self._reportError("Expected file name string or list of file names for output data", **kwargs)

        if OutputSE:
            description = "User specified Output SE"
            if isinstance(OutputSE, str):
                OutputSE = [OutputSE]
            elif not isinstance(OutputSE, list):
                return self._reportError("Expected string or list for OutputSE", **kwargs)
            OutputSE = ";".join(OutputSE)
            self._addParameter(self.workflow, "UserOutputSE", "JDL", OutputSE, description)

        if OutputPath:
            description = "User specified Output Path"
            if not isinstance(OutputPath, str):
                return self._reportError("Expected string for OutputPath", **kwargs)
            # Remove leading "/" that might cause problems with os.path.join
            while OutputPath[0] == "/":
                OutputPath = OutputPath[1:]
            self._addParameter(self.workflow, "UserOutputPath", "JDL", OutputPath, description)

        if replicate:
            self._addParameter(
                self.workflow, "ReplicateUserOutputData", "string", replicate, "Flag to replicate or not"
            )

        if filePrepend:
            keepcharacters = (".", "_")
            prependString = "".join(c for c in filePrepend if c.isalnum() or c in keepcharacters).rstrip()
            self._addParameter(self.workflow, "UserOutputLFNPrepend", "string", prependString, "String to prepend")
            # no '_' in lfns are allowed
            for lfn in lfns:
                if "_" in lfn:
                    self.log.error("When using filePrepend, file names can't contain '_'")
                    return S_ERROR("When using filePrepend, file names can't contain '_'")

        return S_OK()

    #############################################################################

    def setExecutable(
        self, executable, arguments=None, logFile=None, systemConfig="ANY", modulesNameList=None, parametersList=None
    ):
        """Specifies executable script to run with optional arguments and log file
        for standard output.

           These can be either:

            - Submission of a python or shell script to DIRAC
               - Can be inline scripts e.g. C{'/bin/ls'}
               - Scripts as executables e.g. python or shell script file

           Example usage:

           >>> job = Job()
           >>> job.setExecutable('myScript.py')

           :param executable: Executable
           :type executable: string
           :param arguments: Optional arguments to executable
           :type arguments: string
           :param logFile: Optional log file name
           :type logFile: string
        """
        kwargs = {"executable": executable, "arguments": arguments, "applicationLog": logFile}
        if not isinstance(executable, str):
            return self._reportError("Expected strings for executable and arguments", **kwargs)

        if os.path.exists(executable):
            self.log.verbose(f"Found script executable file {executable}")
            self.addToInputSandbox.append(executable)
            logName = f"{os.path.basename(executable)}.log"
            moduleName = os.path.basename(executable)
        else:
            self.log.verbose("Found executable code")
            logName = "CodeOutput.log"
            moduleName = "CodeSegment"

        if logFile:
            if isinstance(logFile, str):
                logName = str(logFile)

        self.stepCount += 1

        moduleName = moduleName.replace(".", "")
        stepName = f"ScriptStep{self.stepCount}"

        if not modulesNameList:
            modulesNameList = ["CreateDataFile", "LHCbScript", "FileUsage", "UserJobFinalization"]
        if not parametersList:
            parametersList = [
                ("name", "string", "", "Name of executable"),
                ("executable", "string", "", "Executable Script"),
                ("arguments", "string", "", "Arguments for executable Script"),
                ("applicationLog", "string", "", "Log file name"),
                ("SystemConfig", "string", "", "binary tag"),
            ]

        step = getStepDefinition(
            stepName,
            modulesNameList=modulesNameList,
            importLine="LHCbDIRAC.Workflow.Modules",
            parametersList=parametersList,
        )

        stepName = f"RunScriptStep{self.stepCount}"
        logPrefix = f"Script{self.stepCount}_"
        logName = f"{logPrefix}{logName}"
        self.addToOutputSandbox.append(logName)

        stepInstance = addStepToWorkflow(self.workflow, step, stepName)

        stepInstance.setValue("name", moduleName)
        stepInstance.setValue("applicationLog", logName)
        stepInstance.setValue("executable", executable)
        stepInstance.setValue("SystemConfig", systemConfig)
        if arguments:
            stepInstance.setValue("arguments", arguments)

        return S_OK()

    #############################################################################

    def setDIRACPlatform(self):
        """Use
        LHCbDIRAC.ConfigurationSystem.Client.Helpers.Resources.getPlatformForJob
        for determining DIRAC platform.

        :returns: S_OK/S_ERROR
        """
        platform = getPlatformForJob(self.workflow)
        if platform:
            return self.setPlatform(platform)
        return S_OK()

    def setPlatform(self, platform):
        """Developer function: sets the target platform, e.g. x86_64-slc6, or
        x86_64-slc6.avx2 This platform is in the form of what it is returned by the
        dirac-architecture script)

        Normally, this method should not be called directly. Instead, clients should call setDIRACPlatfom()

        FIXME: this method is similar (but not same) to what is in Vanilla DIRAC Job.py
               and should be evaluated if to change the base one and remove this.

        :returns: S_OK/S_ERROR
        """
        kwargs = {"platform": platform}

        if not isinstance(platform, str):
            return self._reportError("Expected string for platform", **kwargs)

        if platform and platform.lower() != "any":
            # This is used in JobDB.__checkAndPrepareJob
            self._addParameter(self.workflow, "Platform", "JDL", platform, "Platform (host OS + micro arch)")
        return S_OK()

    #############################################################################

    def setInputData(self, lfns, bkClient=None, runNumber=None, persistencyType=None):
        """Add the input data and the run number, if available."""

        if not lfns:
            self.log.warn("no lfns passed in setInputData, was that intentional?")
            return S_OK("Nothing to do")

        res = Job.setInputData(self, lfns)
        if not res["OK"]:
            return res

        if not runNumber or not persistencyType:
            if not bkClient:
                bkClient = BookkeepingClient()

        if not runNumber:
            res = bkClient.getFileMetadata(lfns)
            if not res["OK"]:
                return res

            runNumbers = {
                fileMeta["RunNumber"]
                for fileMeta in res["Value"]["Successful"].values()
                if isinstance(fileMeta.get("RunNumber"), int) and fileMeta["RunNumber"] > 0
            }

            if len(runNumbers) == 1:
                runNumber = str(runNumbers.pop())

        if runNumber and int(runNumber) > 0:
            self._addParameter(self.workflow, "runNumber", "JDL", runNumber, "Input run number")

        if not persistencyType:
            res = bkClient.getFileTypeVersion(lfns)
            if not res["OK"]:
                return res

            typeVersions = res["Value"]

            if not typeVersions:
                self.log.verbose("The requested files do not exist in the BKK")
            else:
                self.log.verbose(f"Found file types {typeVersions.values()} for LFNs: {list(typeVersions)}")
                typeVersionsList = list(set(typeVersions.values()))
                if len(typeVersionsList) == 1:
                    persistencyType = typeVersionsList[0]

        if persistencyType:
            self._addParameter(
                self.workflow, "persistency", "String", persistencyType, "Persistency type of the inputs"
            )

        return S_OK()

    #############################################################################

    def setRunMetadata(self, runMetadataDict):
        """set the run metadata."""

        self._addParameter(self.workflow, "runMetadata", "String", str(runMetadataDict), "Input run metadata")
        return S_OK()

    #############################################################################

    def runLocal(self, diracLHCb=None):
        """The DiracLHCb (API) object is for local submission.

        A BKKClient might be needed. First, adds Ancestors (in any) to the
        InputData.
        """

        if diracLHCb is not None:
            diracLHCb = diracLHCb
        else:
            diracLHCb = DiracLHCb()

        return diracLHCb.submitJob(self, mode="local")


# EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""stores the data taking conditions."""
from DIRAC import gLogger


class DataTakingConditions:
    """DataTakingConditions class."""

    def __init__(self):
        """initialize the class memeber."""
        self.parameters = {}

    def writeToXML(self):
        """creates an xml string."""
        gLogger.info("Write DataTaking conditions to XML!!")
        result = "<DataTakingConditions>\n"
        for name, value in self.parameters.items():
            result += '    <Parameter Name="' + name + '"   Value="' + value + '"/>\n'
        result += "</DataTakingConditions>\n"

        return result



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""Unit tests for Workflow Module UploadMC."""
# pylint: disable=protected-access, missing-docstring, invalid-name, line-too-long
from itertools import product

from unittest.mock import MagicMock
import pytest

from LHCbDIRAC.Workflow.Modules.mock_Commons import wf_commons, step_commons

# sut
from LHCbDIRAC.Workflow.Modules.UploadMC import UploadMC


def test_instantiation(mocker):
    mocker.patch("LHCbDIRAC.Workflow.Modules.ModuleBase.RequestValidator", side_effect=MagicMock())
    uploadMC = UploadMC()
    assert isinstance(uploadMC, UploadMC)


def test__resolveInputVariables(mocker):
    mocker.patch("LHCbDIRAC.Workflow.Modules.ModuleBase.RequestValidator", side_effect=MagicMock())
    mocker.patch("LHCbDIRAC.Workflow.Modules.ModuleBase.FileReport", side_effect=MagicMock())
    uploadMC = UploadMC()
    uploadMC.workflow_commons = {}
    uploadMC.step_commons = {}
    uploadMC._resolveInputVariables()


allCombinations = list(product(wf_commons + [{}], step_commons + [{}]))


@pytest.mark.parametrize("wf_common, s_common", allCombinations)
def test_execute(mocker, wf_common, s_common):
    mocker.patch("LHCbDIRAC.Workflow.Modules.ModuleBase.RequestValidator", side_effect=MagicMock())
    mocker.patch("LHCbDIRAC.Workflow.Modules.ModuleBase.FileReport", side_effect=MagicMock())
    uploadMC = UploadMC()
    uploadMC.STEP_NUMBER = "1"
    uploadMC.workflow_commons = wf_common
    uploadMC.step_commons = s_common
    uploadMC.execute()



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""Resolve SE takes the workflow SE description and returns the list of
destination storage elements for uploading an output file."""
from DIRAC import gLogger
from DIRAC.DataManagementSystem.Utilities.ResolveSE import getDestinationSEList as diracGetDestinationSEList


def getDestinationSEList(outputSE, site, outputmode="Any", run=None):
    """Evaluate the output SE list from a workflow and return the concrete list
    of SEs to upload output data."""
    if outputmode.lower() not in ("any", "local", "run"):
        raise RuntimeError("Unexpected outputmode")

    if outputmode.lower() == "run":
        gLogger.verbose("Output mode set to 'run', thus ignoring site parameter")
        if not run:
            raise RuntimeError("Expected runNumber")
        try:
            run = int(run)
        except ValueError as ve:
            raise RuntimeError(f"Expected runNumber as a number: {ve}")

        gLogger.debug("RunNumber = %d" % run)
        from LHCbDIRAC.TransformationSystem.Client.TransformationClient import TransformationClient

        runDestination = TransformationClient().getDestinationForRun(run)
        if not runDestination["OK"] or run not in runDestination["Value"]:
            raise RuntimeError(
                "Issue getting destinationForRun (%d): " % run + runDestination.get("Message", "unknown run")
            )
        site = runDestination["Value"][run]
        gLogger.verbose("Site set to %s for run %d" % (site, run))
        outputmode = "Local"

    return diracGetDestinationSEList(outputSE, site, outputmode)
